# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from logilab.common.testlib import unittest_main, mock_object
from logilab.common import tempattr

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.predicates import is_instance
from cubicweb.web.views import uicfg
from cubicweb.web.views.editforms import DeleteConfFormView
from cubicweb.web.formwidgets import AutoCompletionWidget
from cubicweb.schema import RRQLExpression


AFFK = uicfg.autoform_field_kwargs
AFS = uicfg.autoform_section


def rbc(entity, formtype, section):
    if section in ('attributes', 'metadata', 'hidden'):
        permission = 'update'
    else:
        permission = 'add'
    return [(rschema.type, x)
            for rschema, tschemas, x in AFS.relations_by_section(entity,
                                                                 formtype,
                                                                 section,
                                                                 permission)]

class AutomaticEntityFormTC(CubicWebTC):

    def test_custom_widget(self):
        with self.admin_access.web_request() as req:
            AFFK.tag_subject_of(('CWUser', 'login', '*'),
                                {'widget': AutoCompletionWidget(autocomplete_initfunc='get_logins')})
            form = self.vreg['forms'].select('edition', req, entity=req.user)
            field = form.field_by_name('login', 'subject')
            self.assertIsInstance(field.widget, AutoCompletionWidget)
            AFFK.del_rtag('CWUser', 'login', '*', 'subject')


    def test_cwuser_relations_by_category(self):
        with self.admin_access.web_request() as req:
            e = self.vreg['etypes'].etype_class('CWUser')(req)
            # see custom configuration in views.cwuser
            self.assertEqual(rbc(e, 'main', 'attributes'),
                             [('login', 'subject'),
                              ('upassword', 'subject'),
                              ('firstname', 'subject'),
                              ('surname', 'subject'),
                              ('in_group', 'subject'),
                              ])
            self.assertEqual(rbc(e, 'muledit', 'attributes'),
                                  [('login', 'subject'),
                                   ('upassword', 'subject'),
                                   ('in_group', 'subject'),
                                   ])
            self.assertCountEqual(rbc(e, 'main', 'metadata'),
                                  [('last_login_time', 'subject'),
                                   ('cw_source', 'subject'),
                                   ('creation_date', 'subject'),
                                   ('modification_date', 'subject'),
                                   ('created_by', 'subject'),
                                   ('owned_by', 'subject'),
                                   ('bookmarked_by', 'object'),
                                   ])
            # XXX skip 'tags' relation here and in the hidden category because
            # of some test interdependancy when pytest is launched on whole cw
            # (appears here while expected in hidden
            self.assertCountEqual([x for x in rbc(e, 'main', 'relations')
                                   if x != ('tags', 'object')],
                                  [('connait', 'subject'),
                                   ('custom_workflow', 'subject'),
                                   ('primary_email', 'subject'),
                                   ('checked_by', 'object'),
                                   ])
            self.assertListEqual(rbc(e, 'main', 'inlined'),
                                  [('use_email', 'subject'),
                                   ])
            # owned_by is defined both as subject and object relations on CWUser
            self.assertListEqual(sorted(x for x in rbc(e, 'main', 'hidden')
                                         if x != ('tags', 'object')),
                                  sorted([('for_user', 'object'),
                                          ('created_by', 'object'),
                                          ('wf_info_for', 'object'),
                                          ('owned_by', 'object'),
                                          ]))

    def test_inlined_view(self):
        self.assertIn('main_inlined',
                      AFS.etype_get('CWUser', 'use_email', 'subject', 'EmailAddress'))
        self.assertNotIn('main_inlined',
                         AFS.etype_get('CWUser', 'primary_email', 'subject', 'EmailAddress'))
        self.assertIn('main_relations',
                      AFS.etype_get('CWUser', 'primary_email', 'subject', 'EmailAddress'))

    def test_personne_relations_by_category(self):
        with self.admin_access.web_request() as req:
            e = self.vreg['etypes'].etype_class('Personne')(req)
            self.assertListEqual(rbc(e, 'main', 'attributes'),
                                  [('nom', 'subject'),
                                   ('prenom', 'subject'),
                                   ('sexe', 'subject'),
                                   ('promo', 'subject'),
                                   ('titre', 'subject'),
                                   ('ass', 'subject'),
                                   ('web', 'subject'),
                                   ('tel', 'subject'),
                                   ('fax', 'subject'),
                                   ('datenaiss', 'subject'),
                                   ('tzdatenaiss', 'subject'),
                                   ('test', 'subject'),
                                   ('description', 'subject'),
                                   ('salary', 'subject'),
                                   ])
            self.assertListEqual(rbc(e, 'muledit', 'attributes'),
                                  [('nom', 'subject'),
                                   ])
            self.assertCountEqual(rbc(e, 'main', 'metadata'),
                                  [('cw_source', 'subject'),
                                   ('creation_date', 'subject'),
                                   ('modification_date', 'subject'),
                                   ('created_by', 'subject'),
                                   ('owned_by', 'subject'),
                                   ])
            self.assertCountEqual(rbc(e, 'main', 'relations'),
                                  [('travaille', 'subject'),
                                   ('manager', 'object'),
                                   ('connait', 'object'),
                                   ])
            self.assertListEqual(rbc(e, 'main', 'hidden'),
                                  [])

    def test_edition_form(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X LIMIT 1')
            form = self.vreg['forms'].select('edition', req, rset=rset, row=0, col=0)
            # should be also selectable by specifying entity
            self.vreg['forms'].select('edition', req, entity=rset.get_entity(0, 0))
            self.assertFalse(any(f for f in form.fields if f is None))

    def test_edition_form_with_action(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X LIMIT 1')
            form = self.vreg['forms'].select('edition', req, rset=rset, row=0,
                                             col=0, action='my_custom_action')
            self.assertEqual(form.form_action(), 'my_custom_action')

    def test_attribute_add_permissions(self):
        # https://www.cubicweb.org/ticket/4342844
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, u'toto')
            cnx.commit()
        with self.new_access(u'toto').web_request() as req:
            e = self.vreg['etypes'].etype_class('Personne')(req)
            cform = self.vreg['forms'].select('edition', req, entity=e)
            self.assertIn('sexe',
                          [rschema.type
                           for rschema, _ in cform.editable_attributes()])
            with self.new_access(u'toto').repo_cnx() as cnx:
                person_eid = cnx.create_entity('Personne', nom=u'Robert').eid
                cnx.commit()
            person = req.entity_from_eid(person_eid)
            mform = self.vreg['forms'].select('edition', req, entity=person)
            self.assertNotIn('sexe',
                             [rschema.type
                              for rschema, _ in mform.editable_attributes()])

    def test_inlined_relations(self):
        with self.admin_access.web_request() as req:
            with self.temporary_permissions(EmailAddress={'add': ()}):
                autoform = self.vreg['forms'].select('edition', req, entity=req.user)
                self.assertEqual(list(autoform.inlined_form_views()), [])

    def test_inlined_form_views(self):
        # when some relation has + cardinality, and some already linked entities which are not
        # updatable, a link to optionally add a new sub-entity should be displayed, not a sub-form
        # forcing creation of a sub-entity
        from cubicweb.web.views import autoform
        with self.admin_access.web_request() as req:
            req.create_entity('EmailAddress', address=u'admin@cubicweb.org',
                              reverse_use_email=req.user.eid)
            use_email_schema = self.vreg.schema['CWUser'].rdef('use_email')
            with tempattr(use_email_schema, 'cardinality', '+1'):
                with self.temporary_permissions(EmailAddress={'update': ()}):
                    form = self.vreg['forms'].select('edition', req, entity=req.user)
                    formviews = list(form.inlined_form_views())
                    self.assertEqual(len(formviews), 1, formviews)
                    self.assertIsInstance(formviews[0], autoform.InlineAddNewLinkView)
            # though do not introduce regression on entity creation with 1 cardinality relation
            with tempattr(use_email_schema, 'cardinality', '11'):
                user = self.vreg['etypes'].etype_class('CWUser')(req)
                form = self.vreg['forms'].select('edition', req, entity=user)
                formviews = list(form.inlined_form_views())
                self.assertEqual(len(formviews), 1, formviews)
                self.assertIsInstance(formviews[0], autoform.InlineEntityCreationFormView)

    def test_check_inlined_rdef_permissions(self):
        # try to check permissions when creating an entity ('user' below is a
        # fresh entity without an eid)
        with self.admin_access.web_request() as req:
            ttype = 'EmailAddress'
            rschema = self.schema['use_email']
            rdef =  rschema.rdefs[('CWUser', ttype)]
            tschema = self.schema[ttype]
            role = 'subject'
            with self.temporary_permissions((rdef, {'add': ()})):
                user = self.vreg['etypes'].etype_class('CWUser')(req)
                autoform = self.vreg['forms'].select('edition', req, entity=user)
                self.assertFalse(autoform.check_inlined_rdef_permissions(rschema, role,
                                                                         tschema, ttype))
            # we actually don't care about the actual expression,
            # may_have_permission only checks the presence of such expressions
            expr = RRQLExpression('S use_email O')
            with self.temporary_permissions((rdef, {'add': (expr,)})):
                user = self.vreg['etypes'].etype_class('CWUser')(req)
                autoform = self.vreg['forms'].select('edition', req, entity=user)
                self.assertTrue(autoform.check_inlined_rdef_permissions(rschema, role,
                                                                        tschema, ttype))


class FormViewsTC(CubicWebTC):

    def test_delete_conf_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWGroup X')
            self.view('deleteconf', rset, template=None, req=req).source

    def test_delete_conf_formview_composite(self):
        with self.admin_access.cnx() as cnx:
            d1 = cnx.create_entity('Directory', name=u'dtest1')
            d2 = cnx.create_entity('Directory', name=u'dtest2', parent=d1)
            d3 = cnx.create_entity('Directory', name=u'dtest3', parent=d2)
            d4 = cnx.create_entity('Directory', name=u'dtest4', parent=d1)
            for i in range(3):
                cnx.create_entity('Directory', name=u'child%s' % (i,),
                                  parent=d3)
            cnx.commit()

        class DirectoryDeleteView(DeleteConfFormView):
            __select__ = (DeleteConfFormView.__select__ &
                          is_instance('Directory'))
            show_composite = True

        self.vreg['propertyvalues']['navigation.page-size'] = 3
        with self.admin_access.web_request() as req, \
                self.temporary_appobjects(DirectoryDeleteView):
            rset = req.execute('Directory X WHERE X name "dtest1"')
            source = self.view('deleteconf', rset,
                               template=None, req=req).source.decode('utf-8')
            # Show composite object at depth 1
            # Don't display "And more composite entities" since their are equal
            # to page size
            expected = (
                '<li>'
                '<a href="http://testing.fr/cubicweb/directory/%s">dtest1</a>'
                '<ul class="treeview"><li>'
                '<a href="http://testing.fr/cubicweb/directory/%s">dtest4</a>'
                '</li><li class="last">'
                '<a href="http://testing.fr/cubicweb/directory/%s">dtest2</a>'
                '</li></ul></li>') % (d1.eid, d4.eid, d2.eid)
            self.assertIn(expected, source)

            # Page size is reached, show "And more composite entities"
            rset = req.execute('Directory X WHERE X name "dtest3"')
            source = self.view('deleteconf', rset,
                               template=None, req=req).source.decode('utf-8')
            expected = (
                '<li>'
                '<a href="http://testing.fr/cubicweb/directory/%s">dtest3</a>'
                '<ul class="treeview"><li>'
                '<a href="http://testing.fr/cubicweb/directory/%s">child2</a>'
                '</li><li>'
                '<a href="http://testing.fr/cubicweb/directory/%s">child1</a>'
                '</li><li class="last">And more composite entities</li>'
                '</ul></li>') % (
                    d3.eid,
                    req.find('Directory', name='child2').one().eid,
                    req.find('Directory', name='child1').one().eid)
            self.assertIn(expected, source)

    def test_automatic_edition_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X')
            self.view('edition', rset, row=0, template=None, req=req).source

    def test_automatic_edition_copyformview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X')
            self.view('copy', rset, row=0, template=None, req=req).source

    def test_automatic_creation_formview(self):
        with self.admin_access.web_request() as req:
            self.view('creation', None, etype='CWUser', template=None, req=req).source

    def test_automatic_muledit_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X')
            self.view('muledit', rset, template=None, req=req).source

    def test_automatic_reledit_formview(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('CWUser X')
            self.view('reledit', rset, row=0, rtype='login', template=None, req=req).source

    def test_automatic_inline_edit_formview(self):
        with self.admin_access.web_request() as req:
            geid = req.execute('CWGroup X LIMIT 1')[0][0]
            rset = req.execute('CWUser X LIMIT 1')
            self.view('inline-edition', rset, row=0, col=0, rtype='in_group',
                      peid=geid, role='object', i18nctx='', pform=MOCKPFORM,
                      template=None, req=req).source

    def test_automatic_inline_creation_formview(self):
        with self.admin_access.web_request() as req:
            geid = req.execute('CWGroup X LIMIT 1')[0][0]
            self.view('inline-creation', None, etype='CWUser', rtype='in_group',
                      peid=geid, petype='CWGroup', i18nctx='', role='object', pform=MOCKPFORM,
                      template=None, req=req)

MOCKPFORM = mock_object(form_previous_values={}, form_valerror=None)

if __name__ == '__main__':
    unittest_main()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Set of HTML automatic forms to create, delete, copy or edit a single entity
or a list of entities of the same type
"""



from copy import copy

from six.moves import range

from logilab.common.registry import yes
from logilab.common.deprecation import class_moved

from cubicweb import _
from cubicweb import tags
from cubicweb.predicates import (one_line_rset, non_final_entity,
                                 specified_etype_implements, is_instance)
from cubicweb.view import EntityView
from cubicweb.web import stdmsgs, eid_param, formwidgets as fw
from cubicweb.web.form import FormViewMixIn
from cubicweb.web.views import uicfg, forms, reledit

_pvdc = uicfg.primaryview_display_ctrl


class DeleteConfForm(forms.CompositeForm):
    __regid__ = 'deleteconf'
    # XXX non_final_entity does not implement eclass_selector
    __select__ = is_instance('Any')

    domid = 'deleteconf'
    copy_nav_params = True
    form_buttons = [fw.Button(stdmsgs.BUTTON_DELETE, cwaction='delete'),
                    fw.Button(stdmsgs.BUTTON_CANCEL,
                              {'class': fw.Button.css_class + ' cwjs-edition-cancel'})]

    def __init__(self, *args, **kwargs):
        super(DeleteConfForm, self).__init__(*args, **kwargs)
        done = set()
        for entity in self.cw_rset.entities():
            if entity.eid in done:
                continue
            done.add(entity.eid)
            subform = self._cw.vreg['forms'].select('base', self._cw,
                                                    entity=entity,
                                                    mainform=False)
            self.add_subform(subform)


class DeleteConfFormView(FormViewMixIn, EntityView):
    """form used to confirm deletion of some entities"""
    __regid__ = 'deleteconf'
    title = _('delete')
    # don't use navigation, all entities asked to be deleted should be displayed
    # else we will only delete the displayed page
    paginable = False
    # show first level of composite relations in a treeview
    show_composite = False
    show_composite_skip_rtypes = set('wf_info_for',)

    def _iter_composite_entities(self, entity, limit=None):
        for rdef, role in entity.e_schema.composite_rdef_roles:
            if rdef.rtype in self.show_composite_skip_rtypes:
                continue
            for centity in entity.related(
                rdef.rtype, role, limit=limit
            ).entities():
                yield centity

    def call(self, onsubmit=None):
        """ask for confirmation before real deletion"""
        req, w = self._cw, self.w
        _ = req._
        if self.show_composite:
            req.add_css(('jquery-treeview/jquery.treeview.css', 'cubicweb.treeview.css'))
        w(u'<script type="text/javascript">updateMessage(\'%s\');</script>\n'
          % _('this action is not reversible!'))
        # XXX above message should have style of a warning
        w(u'<h4>%s</h4>\n' % _('Do you want to delete the following element(s)?'))
        form = self._cw.vreg['forms'].select(self.__regid__, req,
                                             rset=self.cw_rset,
                                             onsubmit=onsubmit)
        w(u'<ul>\n')
        page_size = req.property_value('navigation.page-size')
        for entity in self.cw_rset.entities():
            # don't use outofcontext view or any other that may contain inline
            # edition form
            w(u'<li>%s' % tags.a(entity.view('textoutofcontext'),
                                 href=entity.absolute_url()))
            if self.show_composite:
                content = None
                for count, centity in enumerate(self._iter_composite_entities(
                    entity, limit=page_size,
                )):
                    if count == 0:
                        w(u'<ul class="treeview">')
                    if content is not None:
                        w(u'<li>%s</li>' % content)
                    if count == page_size - 1:
                        w(u'<li class="last">%s</li></ul>' % _(
                            'And more composite entities'))
                        break
                    content = tags.a(centity.view('textoutofcontext'),
                                     href=centity.absolute_url())
                else:
                    w(u'<li class="last">%s</li></ul>' % content)
            w(u'</li>\n')
        w(u'</ul>\n')
        form.render(w=self.w)


class EditionFormView(FormViewMixIn, EntityView):
    """display primary entity edition form"""
    __regid__ = 'edition'
    # add yes() so it takes precedence over deprecated views in baseforms,
    # though not baseforms based customized view
    __select__ = one_line_rset() & non_final_entity() & yes()
    form_id = 'edition'

    title = _('modification')

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.complete_entity(row, col)
        self.render_form(entity)

    def render_form(self, entity):
        """fetch and render the form"""
        self.form_title(entity)
        form = self._cw.vreg['forms'].select(self.form_id, self._cw,
                                             entity=entity,
                                             submitmsg=self.submited_message())
        self.init_form(form, entity)
        form.render(w=self.w)

    def init_form(self, form, entity):
        """customize your form before rendering here"""
        pass

    def form_title(self, entity):
        """the form view title"""
        ptitle = self._cw._(self.title)
        self.w(u'<div class="formTitle"><span>%s %s</span></div>' % (
            entity.dc_type(), ptitle and '(%s)' % ptitle))

    def submited_message(self):
        """return the message that will be displayed on successful edition"""
        return self._cw._('entity edited')


class CreationFormView(EditionFormView):
    """display primary entity creation form"""
    __regid__ = 'creation'
    __select__ = specified_etype_implements('Any') & yes()

    title = _('creation')

    def call(self, **kwargs):
        """creation view for an entity"""
        # at this point we know etype is a valid entity type, thanks to our
        # selector
        etype = kwargs.pop('etype', self._cw.form.get('etype'))
        entity = self._cw.vreg['etypes'].etype_class(etype)(self._cw)
        entity.eid = next(self._cw.varmaker)
        self.render_form(entity)

    def form_title(self, entity):
        """the form view title"""
        if '__linkto' in self._cw.form:
            if isinstance(self._cw.form['__linkto'], list):
                # XXX which one should be considered (case: add a ticket to a
                # version in jpl)
                rtype, linkto_eid, role = self._cw.form['__linkto'][0].split(':')
            else:
                rtype, linkto_eid, role = self._cw.form['__linkto'].split(':')
            linkto_rset = self._cw.eid_rset(linkto_eid)
            linkto_type = linkto_rset.description[0][0]
            if role == 'subject':
                title = self._cw.__('creating %s (%s %s %s %%(linkto)s)' % (
                    entity.e_schema, entity.e_schema, rtype, linkto_type))
            else:
                title = self._cw.__('creating %s (%s %%(linkto)s %s %s)' % (
                    entity.e_schema, linkto_type, rtype, entity.e_schema))
            msg = title % {'linkto' : self._cw.view('incontext', linkto_rset)}
            self.w(u'<div class="formTitle notransform"><span>%s</span></div>' % msg)
        else:
            super(CreationFormView, self).form_title(entity)

    def url(self):
        """return the url associated with this view"""
        req = self._cw
        return req.vreg["etypes"].etype_class(req.form['etype']).cw_create_url(
            req)

    def submited_message(self):
        """return the message that will be displayed on successful edition"""
        return self._cw._('entity created')


class CopyFormView(EditionFormView):
    """display primary entity creation form initialized with values from another
    entity
    """
    __regid__ = 'copy'

    title = _('copy')
    warning_message = _('Please note that this is only a shallow copy')

    def render_form(self, entity):
        """fetch and render the form"""
        # make a copy of entity to avoid altering the entity in the
        # request's cache.
        entity.complete()
        self.newentity = copy(entity)
        self.copying = entity
        self.newentity.eid = next(self._cw.varmaker)
        self.w(u'<script type="text/javascript">updateMessage("%s");</script>\n'
               % self._cw._(self.warning_message))
        super(CopyFormView, self).render_form(self.newentity)
        del self.newentity

    def init_form(self, form, entity):
        """customize your form before rendering here"""
        super(CopyFormView, self).init_form(form, entity)
        if entity.eid == self.newentity.eid:
            form.add_hidden(eid_param('__cloned_eid', entity.eid),
                            self.copying.eid)
        for rschema, role in form.editable_attributes():
            if not rschema.final:
                # ensure relation cache is filed
                rset = self.copying.related(rschema, role)
                self.newentity.cw_set_relation_cache(rschema, role, rset)

    def submited_message(self):
        """return the message that will be displayed on successful edition"""
        return self._cw._('entity copied')


class TableEditForm(forms.CompositeForm):
    __regid__ = 'muledit'
    domid = 'entityForm'
    onsubmit = "return validateForm('%s', null);" % domid
    form_buttons = [fw.SubmitButton(_('validate modifications on selected items')),
                    fw.ResetButton(_('revert changes'))]

    def __init__(self, req, rset, **kwargs):
        kwargs.setdefault('__redirectrql', rset.printable_rql())
        super(TableEditForm, self).__init__(req, rset=rset, **kwargs)
        for row in range(len(self.cw_rset)):
            form = self._cw.vreg['forms'].select('edition', self._cw,
                                                 rset=self.cw_rset, row=row,
                                                 formtype='muledit',
                                                 copy_nav_params=False,
                                                 mainform=False)
            # XXX rely on the EntityCompositeFormRenderer to put the eid input
            form.remove_field(form.field_by_name('eid'))
            self.add_subform(form)


class TableEditFormView(FormViewMixIn, EntityView):
    __regid__ = 'muledit'
    __select__ = EntityView.__select__ & yes()
    title = _('multiple edit')

    def call(self, **kwargs):
        """a view to edit multiple entities of the same type the first column
        should be the eid
        """
        # XXX overriding formvid (eg __form_id) necessary to make work edition:
        # the edit controller try to select the form with no rset but
        # entity=entity, and use this form to edit the entity. So we want
        # edition form there but specifying formvid may have other undesired
        # side effect. Maybe we should provide another variable optionally
        # telling which form the edit controller should select (eg difffers
        # between html generation / post handling form)
        form = self._cw.vreg['forms'].select(self.__regid__, self._cw,
                                             rset=self.cw_rset,
                                             copy_nav_params=True,
                                             formvid='edition')
        form.render(w=self.w)


# click and edit handling ('reledit') ##########################################

ClickAndEditFormView = class_moved(reledit.ClickAndEditFormView)
AutoClickAndEditFormView = class_moved(reledit.AutoClickAndEditFormView)



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""twisted server configurations:

* the "all-in-one" configuration to get a web instance running in a twisted
  web server integrating a repository server in the same process (only available
  if the repository part of the software is installed
"""


from os.path import join

from logilab.common.configuration import Method, merge_options

from cubicweb.cwconfig import CONFIGURATIONS
from cubicweb.server.serverconfig import ServerConfiguration
from cubicweb.web.webconfig import WebConfiguration


class WebConfigurationBase(WebConfiguration):
    """web instance (in a twisted web server) client of a RQL server"""

    options = merge_options((
        # ctl configuration
        ('port',
         {'type' : 'int',
          'default': None,
          'help': 'http server port number (default to 8080)',
          'group': 'web', 'level': 0,
          }),
        ('interface',
         {'type' : 'string',
          'default': '0.0.0.0',
          'help': 'http server address on which to listen (default to everywhere)',
          'group': 'web', 'level': 1,
          }),
        ('max-post-length',
         {'type' : 'bytes',
          'default': '100MB',
          'help': 'maximum length of HTTP request. Default to 100 MB.',
          'group': 'web', 'level': 1,
          }),
        ('profile',
         {'type' : 'string',
          'default': None,
          'help': 'profile code and use the specified file to store stats if this option is set',
          'group': 'web', 'level': 3,
          }),
        ('host',
         {'type' : 'string',
          'default': None,
          'help': 'host name if not correctly detectable through gethostname',
          'group': 'main', 'level': 1,
          }),
        ('pid-file',
         {'type' : 'string',
          'default': Method('default_pid_file'),
          'help': 'repository\'s pid file',
          'group': 'main', 'level': 2,
          }),
        ('uid',
         {'type' : 'string',
          'default': None,
          'help': 'if this option is set, use the specified user to start \
the repository rather than the user running the command',
          'group': 'main', 'level': WebConfiguration.mode == 'system'
          }),
        ('webserver-threadpool-size',
         {'type': 'int',
          'default': 4,
          'help': "size of twisted's reactor threadpool. It should probably be not too \
much greater than connection-poolsize",
          'group': 'web', 'level': 3,
          }),
        ) + WebConfiguration.options)

    def server_file(self):
        return join(self.apphome, '%s-%s.py' % (self.appid, self.name))

    def default_base_url(self):
        from socket import getfqdn
        return 'http://%s:%s/' % (self['host'] or getfqdn().lower(), self['port'] or 8080)


class AllInOneConfiguration(WebConfigurationBase, ServerConfiguration):
    """repository and web instance in the same twisted process"""
    name = 'all-in-one'
    options = merge_options(WebConfigurationBase.options
                            + ServerConfiguration.options)

    cubicweb_appobject_path = WebConfigurationBase.cubicweb_appobject_path | ServerConfiguration.cubicweb_appobject_path
    cube_appobject_path = WebConfigurationBase.cube_appobject_path | ServerConfiguration.cube_appobject_path


CONFIGURATIONS.append(AllInOneConfiguration)



# copyright 2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unittests for cw.web.formwidgets"""

from logilab.common.testlib import unittest_main, mock_object as mock

from cubicweb.devtools import fake
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web import formwidgets, formfields
from cubicweb.web.views.forms import FieldsForm


class WidgetsTC(CubicWebTC):

    def test_editableurl_widget(self):
        field = formfields.guess_field(self.schema['Bookmark'], self.schema['path'])
        widget = formwidgets.EditableURLWidget()
        req = fake.FakeRequest(form={'path-subjectfqs:A': 'param=value&vid=view'})
        form = mock(_cw=req, formvalues={}, edited_entity=mock(eid='A'))
        self.assertEqual(widget.process_field_data(form, field),
                         '?param=value%26vid%3Dview')

    def test_bitselect_widget(self):
        field = formfields.guess_field(self.schema['CWAttribute'], self.schema['ordernum'])
        field.choices = [('un', '1',), ('deux', '2',)]
        widget = formwidgets.BitSelect()
        req = fake.FakeRequest(form={'ordernum-subject:A': ['1', '2']})
        form = mock(_cw=req, formvalues={}, edited_entity=mock(eid='A'),
                    form_previous_values=())
        self.assertMultiLineEqual(widget._render(form, field, None),
                             '''\
<select id="ordernum-subject:A" multiple="multiple" name="ordernum-subject:A" size="2">
<option selected="selected" value="2">deux</option>
<option selected="selected" value="1">un</option>
</select>''')
        self.assertEqual(widget.process_field_data(form, field),
                         3)

    def test_xml_escape_checkbox(self):
        class TestForm(FieldsForm):
            bool = formfields.BooleanField(ignore_req_params=True,
                choices=[('python >> others', '1')],
                widget=formwidgets.CheckBox())
        with self.admin_access.web_request() as req:
            form = TestForm(req, None)
            form.build_context()
            field = form.field_by_name('bool')
            widget = field.widget
            self.assertMultiLineEqual(widget._render(form, field, None),
                '<label><input id="bool" name="bool" '
                'type="checkbox" value="1" />&#160;'
                'python &gt;&gt; others</label>')


if __name__ == '__main__':
    unittest_main()



# copyright 2003-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Fake objects to ease testing of cubicweb without a fully working environment
"""

from contextlib import contextmanager

from six import string_types

from logilab.database import get_db_helper

from cubicweb.req import RequestSessionBase
from cubicweb.cwvreg import CWRegistryStore
from cubicweb.web.request import ConnectionCubicWebRequestBase

from cubicweb.devtools import BASE_URL, BaseApptestConfiguration


class FakeConfig(dict, BaseApptestConfiguration):
    translations = {}
    uiprops = {}
    apphome = None
    debugmode = False
    def __init__(self, appid='data', apphome=None, cubes=()):
        self.appid = appid
        self.apphome = apphome
        self._cubes = cubes
        self['auth-mode'] = 'cookie'
        self['uid'] = None
        self['base-url'] = BASE_URL
        self['rql-cache-size'] = 3000
        self.datadir_url = BASE_URL + 'data/'

    def cubes(self, expand=False):
        return self._cubes

    def sources(self):
        return {'system': {'db-driver': 'sqlite'}}


class FakeCWRegistryStore(CWRegistryStore):

    def property_value(self, key):
        if key == 'ui.language':
            return 'en'
        assert False


class FakeRequest(ConnectionCubicWebRequestBase):
    """test implementation of an cubicweb request object"""

    def __init__(self, *args, **kwargs):
        if not (args or 'vreg' in kwargs):
            kwargs['vreg'] = FakeCWRegistryStore(FakeConfig(), initlog=False)
        self._http_method = kwargs.pop('method', 'GET')
        self._url = kwargs.pop('url', None)
        if self._url is None:
            self._url = 'view?rql=Blop&vid=blop'
        super(FakeRequest, self).__init__(*args, **kwargs)
        self._session_data = {}

    def set_cookie(self, name, value, maxage=300, expires=None, secure=False, httponly=False):
        super(FakeRequest, self).set_cookie(name, value, maxage, expires, secure, httponly)
        cookie = self.get_response_header('Set-Cookie')
        self._headers_in.setHeader('Cookie', cookie)

    ## Implement request abstract API
    def http_method(self):
        return self._http_method

    def relative_path(self, includeparams=True):
        """return the normalized path of the request (ie at least relative
        to the instance's root, but some other normalization may be needed
        so that the returned path may be used to compare to generated urls
        """
        if self._url.startswith(BASE_URL):
            url = self._url[len(BASE_URL):]
        else:
            url = self._url
        if includeparams:
            return url
        return url.split('?', 1)[0]

    def set_request_header(self, header, value, raw=False):
        """set an incoming HTTP header (for test purpose only)"""
        if isinstance(value, string_types):
            value = [value]
        if raw:
            # adding encoded header is important, else page content
            # will be reconverted back to unicode and apart unefficiency, this
            # may cause decoding problem (e.g. when downloading a file)
            self._headers_in.setRawHeaders(header, value)
        else:
            self._headers_in.setHeader(header, value) #

    def get_response_header(self, header, default=None, raw=False):
        """return output header (for test purpose only)"""
        if raw:
            return self.headers_out.getRawHeaders(header, [default])[0]
        return self.headers_out.getHeader(header, default)

    def build_url_params(self, **kwargs):
        # overriden to get predictable resultts
        args = []
        for param, values in sorted(kwargs.items()):
            if not isinstance(values, (list, tuple)):
                values = (values,)
            for value in values:
                assert value is not None
                args.append(u'%s=%s' % (param, self.url_quote(value)))
        return '&'.join(args)

class FakeUser(object):
    login = 'toto'
    eid = 0
    def in_groups(self, groups):
        return True


class FakeConnection(RequestSessionBase):

    def __init__(self, repo=None, user=None, vreg=None):
        self.repo = repo
        if vreg is None:
            vreg = getattr(self.repo, 'vreg', None)
        if vreg is None:
            vreg = FakeCWRegistryStore(FakeConfig(), initlog=False)
        self.vreg = vreg
        self.cnxset = FakeConnectionsSet()
        self.user = user or FakeUser()
        self.is_internal_session = False
        self.transaction_data = {}

    def execute(self, *args, **kwargs):
        pass

    def commit(self, *args):
        self.transaction_data.clear()

    def system_sql(self, sql, args=None):
        pass

    def set_entity_cache(self, entity):
        pass

    def security_enabled(self, read=False, write=False):
        class FakeCM(object):
            def __enter__(self): pass
            def __exit__(self, exctype, exc, traceback): pass
        return FakeCM()

    # for use with enabled_security context manager
    read_security = write_security = True

    @contextmanager
    def running_hooks_ops(self):
        yield


class FakeRepo(object):
    querier = None

    def __init__(self, schema, vreg=None, config=None):
        self.eids = {}
        self._count = 0
        self.schema = schema
        self.config = config or FakeConfig()
        self.vreg = vreg or FakeCWRegistryStore(self.config, initlog=False)
        self.vreg.schema = schema


class FakeSource(object):
    dbhelper = get_db_helper('sqlite')
    def __init__(self, uri):
        self.uri = uri


class FakeConnectionsSet(object):
    def source(self, uri):
        return FakeSource(uri)



# copyright 2013-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unittest for cubicweb.repoapi"""


from cubicweb.devtools.testlib import CubicWebTC

from cubicweb import ProgrammingError
from cubicweb.repoapi import Connection, connect, anonymous_cnx


class REPOAPITC(CubicWebTC):

    def test_cnx_basic_usage(self):
        """Test that a client connection can be used to access the database"""
        with self.admin_access.client_cnx() as cltcnx:
            # (1) some RQL request
            rset = cltcnx.execute('Any X WHERE X is CWUser')
            self.assertTrue(rset)
            # (2) ORM usage
            random_user = rset.get_entity(0, 0)
            # (3) Write operation
            random_user.cw_set(surname=u'babar')
            # (4) commit
            cltcnx.commit()
            rset = cltcnx.execute('''Any X WHERE X is CWUser,
                                                 X surname "babar"
                                  ''')
            self.assertTrue(rset)
            # prepare test for implicit rollback
            random_user = rset.get_entity(0, 0)
            random_user.cw_set(surname=u'celestine')
        # implicit rollback on exit
        with self.admin_access.client_cnx() as cltcnx:
            rset = cltcnx.execute('''Any X WHERE X is CWUser,
                                                 X surname "babar"
                                  ''')
            self.assertTrue(rset)

    def test_cnx_life_cycle(self):
        """Check that ClientConnection requires explicit open and close
        """
        access = self.admin_access
        cltcnx = Connection(access._repo, access._user)
        # connection not open yet
        with self.assertRaises(ProgrammingError):
            cltcnx.execute('Any X WHERE X is CWUser')
        # connection open and working
        with cltcnx:
            cltcnx.execute('Any X WHERE X is CWUser')
        # connection closed
        with self.assertRaises(ProgrammingError):
            cltcnx.execute('Any X WHERE X is CWUser')

    def test_connect(self):
        """check that repoapi.connect works and returns a usable connection"""
        cnx = connect(self.repo, login='admin', password='gingkow')
        self.assertEqual('admin', cnx.user.login)
        with cnx:
            rset = cnx.execute('Any X WHERE X is CWUser')
            self.assertTrue(rset)

    def test_anonymous_connect(self):
        """check that you can get anonymous connection when the data exist"""
        cnx = anonymous_cnx(self.repo)
        self.assertEqual('anon', cnx.user.login)
        with cnx:
            rset = cnx.execute('Any X WHERE X is CWUser')
            self.assertTrue(rset)

    def test_cnx_has_lang(self):
        """check that client and repo cnx have .lang set"""
        with self.admin_access.client_cnx() as cnx:
            self.assertEqual(cnx.lang, 'en')
        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx.lang, 'en')


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Twisted request handler for CubicWeb"""

from six import text_type

from cubicweb.web.request import CubicWebRequestBase


class CubicWebTwistedRequestAdapter(CubicWebRequestBase):
    """ from twisted .req to cubicweb .form
    req.files are put into .form[<filefield>]
    """
    def __init__(self, req, vreg):
        self._twreq = req
        super(CubicWebTwistedRequestAdapter, self).__init__(
            vreg, req.args, headers=req.received_headers)
        for key, name_stream_list in req.files.items():
            for name, stream in name_stream_list:
                if name is not None:
                    name = text_type(name, self.encoding)
                self.form.setdefault(key, []).append((name, stream))
            # 3.16.4 backward compat
            if len(self.form[key]) == 1:
                self.form[key] = self.form[key][0]
        self.content = self._twreq.content  # stream

    def http_method(self):
        """returns 'POST', 'GET', 'HEAD', etc."""
        return self._twreq.method

    def relative_path(self, includeparams=True):
        """return the normalized path of the request (ie at least relative to
        the instance's root, but some other normalization may be needed so that
        the returned path may be used to compare to generated urls

        :param includeparams:
           boolean indicating if GET form parameters should be kept in the path
        """
        path = self._twreq.uri[1:]  # remove the root '/'
        if not includeparams:
            path = path.split('?', 1)[0]
        return path



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from logilab.common.testlib import unittest_main
from logilab.mtconverter import html_unescape

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.utils import json
from cubicweb.view import StartupView, TRANSITIONAL_DOCTYPE
from cubicweb.web.views import vid_from_rset

def loadjson(value):
    return json.loads(html_unescape(value))

class VidFromRsetTC(CubicWebTC):

    def test_no_rset(self):
        with self.admin_access.web_request() as req:
            self.assertEqual(vid_from_rset(req, None, self.schema), 'index')

    def test_no_entity(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X login "blabla"')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'noresult')

    def test_one_entity(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X login "admin"')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'primary')
            rset = req.execute('Any X, L WHERE X login "admin", X login L')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'primary')
            req.search_state = ('pasnormal',)
            rset = req.execute('Any X WHERE X login "admin"')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'outofcontext-search')

    def test_one_entity_eid(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X eid 1')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'primary')

    def test_more_than_one_entity_same_type(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X is CWUser')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'sameetypelist')
            rset = req.execute('Any X, L WHERE X login L')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'sameetypelist')

    def test_more_than_one_entity_diff_type(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X is IN (CWUser, CWGroup)')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'list')

    def test_more_than_one_entity_by_row(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X, G WHERE X in_group G')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'table')

    def test_more_than_one_entity_by_row_2(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X, GN WHERE X in_group G, G name GN')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'table')

    def test_aggregat(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X, COUNT(T) GROUPBY X WHERE X is T')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'table')
            rset = req.execute('Any MAX(X) WHERE X is CWUser')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'table')

    def test_subquery(self):
        with self.admin_access.web_request() as req:
            rset = req.execute(
'DISTINCT Any X,N ORDERBY N '
'WITH X,N BEING ('
'     (DISTINCT Any P,N WHERE P is CWUser, P login N)'
'       UNION'
'     (DISTINCT Any W,N WHERE W is CWGroup, W name N))')
            self.assertEqual(vid_from_rset(req, rset, self.schema), 'table')


class TableViewTC(CubicWebTC):

    def _prepare_entity(self, req):
        e = req.create_entity("State", name=u'<toto>', description=u'loo"ong blabla')
        rset = req.execute('Any X, D, CD, NOW - CD WHERE X is State, '
                           'X description D, X creation_date CD, X eid %(x)s',
                           {'x': e.eid})
        view = self.vreg['views'].select('table', req, rset=rset)
        return e, rset, view

    def test_sortvalue(self):
        with self.admin_access.web_request() as req:
            e, _, view = self._prepare_entity(req)
            colrenderers = view.build_column_renderers()[:3]
            self.assertListEqual([renderer.sortvalue(0) for renderer in colrenderers],
                                 [u'<toto>', u'loo"ong blabla', e.creation_date])


class HTMLStreamTests(CubicWebTC):

    def test_set_doctype_reset_xmldecl(self):
        """
        tests `cubicweb.web.request.CubicWebRequestBase.set_doctype`
        with xmldecl reset
        """
        class MyView(StartupView):
            __regid__ = 'my-view'
            def call(self):
                self._cw.set_doctype('<!DOCTYPE html>')

        with self.admin_access.web_request() as req:
            with self.temporary_appobjects(MyView):
                html_source = self.view('my-view', req=req).source
                source_lines = [line.strip()
                                for line in html_source.splitlines(False)
                                if line.strip()]
                self.assertListEqual([b'<!DOCTYPE html>',
                                      b'<html xmlns:cubicweb="http://www.cubicweb.org" lang="en">'],
                                     source_lines[:2])

    def test_set_doctype_no_reset_xmldecl(self):
        """
        tests `cubicweb.web.request.CubicWebRequestBase.set_doctype`
        with no xmldecl reset
        """
        html_doctype = TRANSITIONAL_DOCTYPE.strip()
        class MyView(StartupView):
            __regid__ = 'my-view'
            def call(self):
                self._cw.set_doctype(html_doctype)
                self._cw.main_stream.set_htmlattrs([('lang', 'cz')])

        with self.admin_access.web_request() as req:
            with self.temporary_appobjects(MyView):
                html_source = self.view('my-view', req=req).source
                source_lines = [line.strip()
                                for line in html_source.splitlines(False)
                                if line.strip()]
                self.assertListEqual([html_doctype.encode('ascii'),
                                      b'<html xmlns:cubicweb="http://www.cubicweb.org" lang="cz">',
                                      b'<head>'],
                                     source_lines[:3])

class BaseViewsTC(CubicWebTC):

    def test_null(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any X WHERE X login "admin"')
            result = req.view('null', rset)
            self.assertEqual(result, u'')

    def test_final(self):
        with self.admin_access.web_request() as req:
            rset = req.execute('Any "<script></script>"')
            result = req.view('final', rset)
            self.assertEqual(result, u'&lt;script&gt;&lt;/script&gt;')

    def test_incontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            result = entity.view('incontext')
            expected = (u'<a href="http://testing.fr/cubicweb/%d" title="">'
                        u'&lt;script&gt;&lt;/script&gt;</a>' % entity.eid)
            self.assertEqual(result, expected)

    def test_outofcontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            result = entity.view('outofcontext')
            expect = (u'<a href="http://testing.fr/cubicweb/%d" title="">'
                      u'&lt;script&gt;&lt;/script&gt;</a>' % entity.eid)
            self.assertEqual(result, expect)

    def test_outofcontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            result = entity.view('oneline')
            expect = (u'<a href="http://testing.fr/cubicweb/%d" title="">'
                      u'&lt;script&gt;&lt;/script&gt;</a>' % entity.eid)
            self.assertEqual(result, expect)

    def test_text(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            result = entity.view('text')
            self.assertEqual(result, u'<script></script>')

    def test_textincontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            result = entity.view('textincontext')
            self.assertEqual(result, u'<script></script>')

    def test_textoutofcontext(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            result = entity.view('textoutofcontext')
            self.assertEqual(result, u'<script></script>')

    def test_list(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            rset = req.execute('Any X WHERE X is CWUser')
            result = req.view('list', rset)
            expected = u'''<ul class="section">
<li><a href="http://testing.fr/cubicweb/%d" title="">&lt;script&gt;&lt;/script&gt;</a></li>
<li><a href="http://testing.fr/cubicweb/cwuser/admin" title="">admin</a></li>
<li><a href="http://testing.fr/cubicweb/cwuser/anon" title="">anon</a></li>
</ul>
''' % entity.eid
            self.assertEqual(result, expected)

    def test_simplelist(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            rset = req.execute('Any X WHERE X is CWUser')
            result = req.view('simplelist', rset)
            expected = (
                u'<div class="section">'
                u'<a href="http://testing.fr/cubicweb/%d" title="">'
                u'&lt;script&gt;&lt;/script&gt;</a></div>'
                u'<div class="section">'
                u'<a href="http://testing.fr/cubicweb/cwuser/admin" title="">admin</a></div>'
                u'<div class="section">'
                u'<a href="http://testing.fr/cubicweb/cwuser/anon" title="">anon</a></div>'
                % entity.eid
            )
            self.assertEqual(result, expected)

    def test_sameetypelist(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            rset = req.execute('Any X WHERE X is CWUser')
            result = req.view('sameetypelist', rset)
            expected = (
                u'<h1>CWUser_plural</h1>'
                u'<div class="section">'
                u'<a href="http://testing.fr/cubicweb/%d" title="">'
                u'&lt;script&gt;&lt;/script&gt;</a></div>'
                u'<div class="section">'
                u'<a href="http://testing.fr/cubicweb/cwuser/admin" title="">admin</a></div>'
                u'<div class="section">'
                u'<a href="http://testing.fr/cubicweb/cwuser/anon" title="">anon</a></div>'
                % entity.eid
            )
            self.assertEqual(expected, result)

    def test_sameetypelist(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            rset = req.execute('Any X WHERE X is CWUser')
            result = req.view('csv', rset)
            expected = (
                u'<a href="http://testing.fr/cubicweb/%d" title="">&lt;script&gt;&lt;/script&gt;</a>, '
                u'<a href="http://testing.fr/cubicweb/cwuser/admin" title="">admin</a>, '
                u'<a href="http://testing.fr/cubicweb/cwuser/anon" title="">anon</a>'
                % entity.eid
            )
            self.assertEqual(result, expected)

    def test_metadata(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWUser', login=u'<script></script>', upassword=u'toto')
            entity.cw_set(creation_date=u'2000-01-01 00:00:00')
            entity.cw_set(modification_date=u'2015-01-01 00:00:00')
            result = entity.view('metadata')
            expected = (
                u'<div>CWUser #%d - <span>latest update on</span>'
                u' <span class="value">2015/01/01</span>,'
                u' <span>created on</span>'
                u' <span class="value">2000/01/01</span></div>'
                % entity.eid
            )
            self.assertEqual(result, expected)


if __name__ == '__main__':
    unittest_main()



from __future__ import absolute_import

from cubicweb.pyramid.rest_api import EntityResource
from cubicweb.pyramid.core import CubicWebPyramidRequest
from pyramid.view import view_config

from cubicweb.pyramid.test import PyramidCWTest


class RestApiTest(PyramidCWTest):
    def includeme(self, config):
        config.include('cubicweb.pyramid.rest_api')
        config.include('cubicweb.pyramid.test.test_rest_api')

    def test_delete(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('CWGroup', name=u'tmp')
            cnx.commit()

        self.login()
        res = self.webapp.delete('/cwgroup/tmp')
        self.assertEqual(res.status_int, 204)

        with self.admin_access.repo_cnx() as cnx:
            self.assertEqual(cnx.find('CWGroup', name=u'tmp').rowcount, 0)

    def test_rql_execute(self):
        with self.admin_access.repo_cnx() as cnx:
            cnx.create_entity('CWGroup', name=u'tmp')
            cnx.commit()
        self.login()
        params = {'test_rql_execute': 'test'}
        self.webapp.get('/cwgroup/tmp', params=params)


@view_config(
    route_name='cwentities',
    context=EntityResource,
    request_method='GET',
    request_param=('test_rql_execute',)
)
def rql_execute_view(context, request):
    """Return 500 response if rset.req is not a CubicWeb request.
    """
    if isinstance(context.rset.req, CubicWebPyramidRequest):
        request.response.status_int = 204
    else:
        request.response.status_int = 500
        request.response.text = 'rset.req is not a CubicWeb request'
    return request.response


def includeme(config):
    config.scan(__name__)


if __name__ == '__main__':
    from unittest import main
    main()



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""rest publishing functions

contains some functions and setup of docutils for cubicweb. Provides the
following ReST directives:

* `eid`, create link to entity in the repository by their eid

* `card`, create link to card entity in the repository by their wikiid
  (proposing to create it when the refered card doesn't exist yet)

* `winclude`, reference to a web documentation file (in wdoc/ directories)

* `sourcecode` (if pygments is installed), source code colorization

* `rql-table`, create a table from a RQL query

"""


import sys
from itertools import chain
from logging import getLogger
from os.path import join

from six import text_type
from six.moves.urllib.parse import urlsplit

from docutils import statemachine, nodes, utils, io
from docutils.core import Publisher
from docutils.parsers.rst import Parser, states, directives, Directive
from docutils.parsers.rst.roles import register_canonical_role, set_classes

from logilab.mtconverter import ESC_UCAR_TABLE, ESC_CAR_TABLE, xml_escape

from cubicweb import UnknownEid
from cubicweb.ext.html4zope import Writer

from cubicweb.web.views import vid_from_rset  # XXX better not to import c.w.views here...

# We provide our own parser as an attempt to get rid of
# state machine reinstanciation

import re
# compile states.Body patterns
for k, v in states.Body.patterns.items():
    if isinstance(v, str):
        states.Body.patterns[k] = re.compile(v)

# register ReStructured Text mimetype / extensions
import mimetypes
mimetypes.add_type('text/rest', '.rest')
mimetypes.add_type('text/rest', '.rst')


LOGGER = getLogger('cubicweb.rest')


def eid_reference_role(role, rawtext, text, lineno, inliner,
                       options={}, content=[]):
    try:
        try:
            eid_num, rest = text.split(u':', 1)
        except ValueError:
            eid_num, rest = text, '#'+text
        eid_num = int(eid_num)
        if eid_num < 0:
            raise ValueError
    except ValueError:
        msg = inliner.reporter.error(
            'EID number must be a positive number; "%s" is invalid.'
            % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    # Base URL mainly used by inliner.pep_reference; so this is correct:
    context = inliner.document.settings.context
    try:
        refedentity = context._cw.entity_from_eid(eid_num)
    except UnknownEid:
        ref = '#'
        rest += u' ' + context._cw._('(UNEXISTANT EID)')
    else:
        ref = refedentity.absolute_url()
    set_classes(options)
    return [nodes.reference(rawtext, utils.unescape(rest), refuri=ref,
                            **options)], []


def rql_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """``:rql:`<rql-expr>``` or ``:rql:`<rql-expr>:<vid>```

    Example: ``:rql:`Any X,Y WHERE X is CWUser, X login Y:table```

    Replace the directive with the output of applying the view to the resultset
    returned by the query.

    "X eid %(userid)s" can be used in the RQL query for this query will be
    executed with the argument {'userid': _cw.user.eid}.
    """
    _cw = inliner.document.settings.context._cw
    text = text.strip()
    if ':' in text:
        rql, vid = text.rsplit(u':', 1)
        rql = rql.strip()
    else:
        rql, vid = text, None
    _cw.ensure_ro_rql(rql)
    try:
        rset = _cw.execute(rql, {'userid': _cw.user.eid})
        if rset:
            if vid is None:
                vid = vid_from_rset(_cw, rset, _cw.vreg.schema)
        else:
            vid = 'noresult'
        view = _cw.vreg['views'].select(vid, _cw, rset=rset)
        content = view.render()
    except Exception as exc:
        content = 'an error occurred while interpreting this rql directive: %r' % exc
    set_classes(options)
    return [nodes.raw('', content, format='html')], []


def bookmark_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    """``:bookmark:`<bookmark-eid>``` or ``:bookmark:`<eid>:<vid>```

    Example: ``:bookmark:`1234:table```

    Replace the directive with the output of applying the view to the resultset
    returned by the query stored in the bookmark. By default, the view is the one
    stored in the bookmark, but it can be overridden by the directive as in the
    example above.

    "X eid %(userid)s" can be used in the RQL query stored in the Bookmark, for
    this query will be executed with the argument {'userid': _cw.user.eid}.
    """
    _cw = inliner.document.settings.context._cw
    text = text.strip()
    try:
        if ':' in text:
            eid, vid = text.rsplit(u':', 1)
            eid = int(eid)
        else:
            eid, vid = int(text), None
    except ValueError:
        msg = inliner.reporter.error(
            'EID number must be a positive number; "%s" is invalid.'
            % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    try:
        bookmark = _cw.entity_from_eid(eid)
    except UnknownEid:
        msg = inliner.reporter.error('Unknown EID %s.' % text, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    try:
        params = dict(_cw.url_parse_qsl(urlsplit(bookmark.path).query))
        rql = params['rql']
        if vid is None:
            vid = params.get('vid')
    except (ValueError, KeyError) as exc:
        msg = inliner.reporter.error('Could not parse bookmark path %s [%s].'
                                     % (bookmark.path, exc), line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    try:
        rset = _cw.execute(rql, {'userid': _cw.user.eid})
        if rset:
            if vid is None:
                vid = vid_from_rset(_cw, rset, _cw.vreg.schema)
        else:
            vid = 'noresult'
        view = _cw.vreg['views'].select(vid, _cw, rset=rset)
        content = view.render()
    except Exception as exc:
        content = 'An error occurred while interpreting directive bookmark: %r' % exc
    set_classes(options)
    return [nodes.raw('', content, format='html')], []


def winclude_directive(name, arguments, options, content, lineno,
                       content_offset, block_text, state, state_machine):
    """Include a reST file as part of the content of this reST file.

    same as standard include directive but using config.locate_doc_resource to
    get actual file to include.

    Most part of this implementation is copied from `include` directive defined
    in `docutils.parsers.rst.directives.misc`
    """
    context = state.document.settings.context
    cw = context._cw
    source = state_machine.input_lines.source(
        lineno - state_machine.input_offset - 1)
    #source_dir = os.path.dirname(os.path.abspath(source))
    fid = arguments[0]
    for lang in chain((cw.lang, cw.vreg.property_value('ui.language')),
                      cw.vreg.config.available_languages()):
        rid = '%s_%s.rst' % (fid, lang)
        resourcedir = cw.vreg.config.locate_doc_file(rid)
        if resourcedir:
            break
    else:
        severe = state_machine.reporter.severe(
              'Problems with "%s" directive path:\nno resource matching %s.'
              % (name, fid),
              nodes.literal_block(block_text, block_text), line=lineno)
        return [severe]
    path = join(resourcedir, rid)
    encoding = options.get('encoding', state.document.settings.input_encoding)
    try:
        state.document.settings.record_dependencies.add(path)
        include_file = io.FileInput(
            source_path=path, encoding=encoding,
            error_handler=state.document.settings.input_encoding_error_handler,
            handle_io_errors=None)
    except IOError as error:
        severe = state_machine.reporter.severe(
              'Problems with "%s" directive path:\n%s: %s.'
              % (name, error.__class__.__name__, error),
              nodes.literal_block(block_text, block_text), line=lineno)
        return [severe]
    try:
        include_text = include_file.read()
    except UnicodeError as error:
        severe = state_machine.reporter.severe(
              'Problem with "%s" directive:\n%s: %s'
              % (name, error.__class__.__name__, error),
              nodes.literal_block(block_text, block_text), line=lineno)
        return [severe]
    if 'literal' in options:
        literal_block = nodes.literal_block(include_text, include_text,
                                            source=path)
        literal_block.line = 1
        return literal_block
    else:
        include_lines = statemachine.string2lines(include_text,
                                                  convert_whitespace=1)
        state_machine.insert_input(include_lines, path)
        return []

winclude_directive.arguments = (1, 0, 1)
winclude_directive.options = {'literal': directives.flag,
                              'encoding': directives.encoding}


class RQLTableDirective(Directive):
    """rql-table directive

    Example:

        .. rql-table::
           :vid: mytable
           :headers: , , progress
           :colvids: 2=progress

            Any X,U,X WHERE X is Project, X url U

    All fields but the RQL string are optionnal. The ``:headers:`` option can
    contain empty column names.
    """

    required_arguments = 0
    optional_arguments = 0
    has_content= True
    final_argument_whitespace = True
    option_spec = {'vid': directives.unchanged,
                   'headers': directives.unchanged,
                   'colvids': directives.unchanged}

    def run(self):
        errid = "rql-table directive"
        self.assert_has_content()
        if self.arguments:
            raise self.warning('%s does not accept arguments' % errid)
        rql = ' '.join([l.strip() for l in self.content])
        _cw = self.state.document.settings.context._cw
        _cw.ensure_ro_rql(rql)
        try:
            rset = _cw.execute(rql)
        except Exception as exc:
            raise self.error("fail to execute RQL query in %s: %r" %
                             (errid, exc))
        if not rset:
            raise self.warning("empty result set")
        vid = self.options.get('vid', 'table')
        try:
            view = _cw.vreg['views'].select(vid, _cw, rset=rset)
        except Exception as exc:
            raise self.error("fail to select '%s' view in %s: %r" %
                             (vid, errid, exc))
        headers = None
        if 'headers' in self.options:
            headers = [h.strip() for h in self.options['headers'].split(',')]
            while headers.count(''):
                headers[headers.index('')] = None
            if len(headers) != len(rset[0]):
                raise self.error("the number of 'headers' does not match the "
                                 "number of columns in %s" % errid)
        cellvids = None
        if 'colvids' in self.options:
            cellvids = {}
            for f in self.options['colvids'].split(','):
                try:
                    idx, vid = f.strip().split('=')
                except ValueError:
                    raise self.error("malformatted 'colvids' option in %s" %
                                     errid)
                cellvids[int(idx.strip())] = vid.strip()
        try:
            content = view.render(headers=headers, cellvids=cellvids)
        except Exception as exc:
            raise self.error("Error rendering %s (%s)" % (errid, exc))
        return [nodes.raw('', content, format='html')]


try:
    from pygments import highlight
    from pygments.lexers import get_lexer_by_name
    from pygments.formatters.html import HtmlFormatter
except ImportError:
    pygments_directive = None
else:
    _PYGMENTS_FORMATTER = HtmlFormatter()

    def pygments_directive(name, arguments, options, content, lineno,
                           content_offset, block_text, state, state_machine):
        try:
            lexer = get_lexer_by_name(arguments[0])
        except ValueError:
            # no lexer found
            lexer = get_lexer_by_name('text')
        parsed = highlight(u'\n'.join(content), lexer, _PYGMENTS_FORMATTER)
        # don't fail if no context set on the sourcecode directive
        try:
            context = state.document.settings.context
            context._cw.add_css('pygments.css')
        except AttributeError:
            # used outside cubicweb XXX use hasattr instead
            pass
        return [nodes.raw('', parsed, format='html')]

    pygments_directive.arguments = (1, 0, 1)
    pygments_directive.content = 1


class CubicWebReSTParser(Parser):
    """The (customized) reStructuredText parser."""

    def __init__(self):
        self.initial_state = 'Body'
        self.state_classes = states.state_classes
        self.inliner = states.Inliner()
        self.statemachine = states.RSTStateMachine(
              state_classes=self.state_classes,
              initial_state=self.initial_state,
              debug=0)

    def parse(self, inputstring, document):
        """Parse `inputstring` and populate `document`, a document tree."""
        self.setup_parse(inputstring, document)
        inputlines = statemachine.string2lines(inputstring,
                                               convert_whitespace=1)
        self.statemachine.run(inputlines, document, inliner=self.inliner)
        self.finish_parse()


# XXX docutils keep a ref on context, can't find a correct way to remove it
class CWReSTPublisher(Publisher):
    def __init__(self, context, settings, **kwargs):
        Publisher.__init__(self, **kwargs)
        self.set_components('standalone', 'restructuredtext', 'pseudoxml')
        self.process_programmatic_settings(None, settings, None)
        self.settings.context = context


def rest_publish(context, data):
    """publish a string formatted as ReStructured Text to HTML

    :type context: a cubicweb application object

    :type data: str
    :param data: some ReST text

    :rtype: unicode
    :return:
      the data formatted as HTML or the original data if an error occurred
    """
    req = context._cw
    if isinstance(data, text_type):
        encoding = 'unicode'
        # remove unprintable characters unauthorized in xml
        data = data.translate(ESC_UCAR_TABLE)
    else:
        encoding = req.encoding
        # remove unprintable characters unauthorized in xml
        data = data.translate(ESC_CAR_TABLE)
    settings = {'input_encoding': encoding, 'output_encoding': 'unicode',
                'warning_stream': False,
                'traceback': True, # don't sys.exit
                'stylesheet': None, # don't try to embed stylesheet (may cause
                                    # obscure bug due to docutils computing
                                    # relative path according to the directory
                                    # used *at import time*
                # dunno what's the max, severe is 4, and we never want a crash
                # (though try/except may be a better option...). May be the
                # above traceback option will avoid this?
                'halt_level': 10,
                # disable stupid switch to colspan=2 if field name is above a size limit
                'field_name_limit': sys.maxsize,
                }
    if context:
        if hasattr(req, 'url'):
            base_url = req.url()
        elif hasattr(context, 'absolute_url'):
            base_url = context.absolute_url()
        else:
            base_url = req.base_url()
    else:
        base_url = None
    try:
        pub = CWReSTPublisher(context, settings,
                              parser=CubicWebReSTParser(),
                              writer=Writer(base_url=base_url),
                              source_class=io.StringInput,
                              destination_class=io.StringOutput)
        pub.set_source(data)
        pub.set_destination()
        res = pub.publish(enable_exit_status=None)
        # necessary for proper garbage collection, else a ref is kept somewhere in docutils...
        del pub.settings.context
        return res
    except BaseException:
        LOGGER.exception('error while publishing ReST text')
        if not isinstance(data, text_type):
            data = text_type(data, encoding, 'replace')
        return xml_escape(req._('error while publishing ReST text')
                           + '\n\n' + data)


_INITIALIZED = False
def cw_rest_init():
    global _INITIALIZED
    if _INITIALIZED:
        return
    _INITIALIZED = True
    register_canonical_role('eid', eid_reference_role)
    register_canonical_role('rql', rql_role)
    register_canonical_role('bookmark', bookmark_role)
    directives.register_directive('winclude', winclude_directive)
    if pygments_directive is not None:
        directives.register_directive('sourcecode', pygments_directive)
    directives.register_directive('rql-table', RQLTableDirective)



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""unit tests for cubicweb.web.application"""

import base64

from six import text_type
from six.moves import http_client
from six.moves.http_cookies import SimpleCookie

from logilab.common.testlib import TestCase, unittest_main
from logilab.common.decorators import clear_cache

from cubicweb import view
from cubicweb.devtools.testlib import CubicWebTC, real_error_handling
from cubicweb.devtools.fake import FakeRequest
from cubicweb.web import LogOut, Redirect, INTERNAL_FIELD_VALUE
from cubicweb.web.views.basecontrollers import ViewController
from cubicweb.web.application import anonymized_request


class FakeMapping:
    """emulates a mapping module"""
    def __init__(self):
        self.ENTITIES_MAP = {}
        self.ATTRIBUTES_MAP = {}
        self.RELATIONS_MAP = {}


class MockCursor:
    def __init__(self):
        self.executed = []

    def execute(self, rql, args=None, build_descr=False):
        args = args or {}
        self.executed.append(rql % args)


class FakeController(ViewController):

    def __init__(self, form=None):
        self._cw = FakeRequest()
        self._cw.form = form or {}
        self._cursor = MockCursor()
        self._cw.execute = self._cursor.execute

    def new_cursor(self):
        self._cursor = MockCursor()
        self._cw.execute = self._cursor.execute

    def set_form(self, form):
        self._cw.form = form


class RequestBaseTC(TestCase):
    def setUp(self):
        self._cw = FakeRequest()

    def test_list_arg(self):
        """tests the list_arg() function"""
        list_arg = self._cw.list_form_param
        self.assertEqual(list_arg('arg3', {}), [])
        d = {'arg1': "value1",
             'arg2': ('foo', INTERNAL_FIELD_VALUE,),
             'arg3': ['bar']}
        self.assertEqual(list_arg('arg1', d, True), ['value1'])
        self.assertEqual(d, {'arg2': ('foo', INTERNAL_FIELD_VALUE), 'arg3': ['bar']})
        self.assertEqual(list_arg('arg2', d, True), ['foo'])
        self.assertEqual({'arg3': ['bar']}, d)
        self.assertEqual(list_arg('arg3', d), ['bar'])
        self.assertEqual({'arg3': ['bar']}, d)

    def test_from_controller(self):
        self._cw.vreg['controllers'] = {'view': 1, 'login': 1}
        self.assertEqual(self._cw.from_controller(), 'view')
        req = FakeRequest(url='project?vid=list')
        req.vreg['controllers'] = {'view': 1, 'login': 1}
        # this assertion is just to make sure that relative_path can be
        # correctly computed as it is used in from_controller()
        self.assertEqual(req.relative_path(False), 'project')
        self.assertEqual(req.from_controller(), 'view')
        # test on a valid non-view controller
        req = FakeRequest(url='login?x=1&y=2')
        req.vreg['controllers'] = {'view': 1, 'login': 1}
        self.assertEqual(req.relative_path(False), 'login')
        self.assertEqual(req.from_controller(), 'login')


class UtilsTC(TestCase):
    """test suite for misc application utilities"""

    def setUp(self):
        self.ctrl = FakeController()

    def test_execute_linkto(self):
        """tests the execute_linkto() function"""
        self.assertEqual(self.ctrl.execute_linkto(), None)
        self.assertEqual(self.ctrl._cursor.executed,
                         [])

        self.ctrl.set_form({'__linkto': 'works_for:12_13_14:object',
                            'eid': 8})
        self.ctrl.execute_linkto()
        self.assertEqual(self.ctrl._cursor.executed,
                         ['SET Y works_for X WHERE X eid 8, Y eid %s' % i
                          for i in (12, 13, 14)])

        self.ctrl.new_cursor()
        self.ctrl.set_form({'__linkto': 'works_for:12_13_14:subject',
                            'eid': 8})
        self.ctrl.execute_linkto()
        self.assertEqual(self.ctrl._cursor.executed,
                         ['SET X works_for Y WHERE X eid 8, Y eid %s' % i
                          for i in (12, 13, 14)])

        self.ctrl.new_cursor()
        self.ctrl._cw.form = {'__linkto': 'works_for:12_13_14:object'}
        self.ctrl.execute_linkto(eid=8)
        self.assertEqual(self.ctrl._cursor.executed,
                         ['SET Y works_for X WHERE X eid 8, Y eid %s' % i
                          for i in (12, 13, 14)])

        self.ctrl.new_cursor()
        self.ctrl.set_form({'__linkto': 'works_for:12_13_14:subject'})
        self.ctrl.execute_linkto(eid=8)
        self.assertEqual(self.ctrl._cursor.executed,
                         ['SET X works_for Y WHERE X eid 8, Y eid %s' % i
                          for i in (12, 13, 14)])


class ApplicationTC(CubicWebTC):

    @classmethod
    def setUpClass(cls):
        super(ApplicationTC, cls).setUpClass()
        cls.config.global_set_option('allow-email-login', True)

    def test_cnx_user_groups_sync(self):
        with self.admin_access.client_cnx() as cnx:
            user = cnx.user
            self.assertEqual(user.groups, set(('managers',)))
            cnx.execute('SET X in_group G WHERE X eid %s, G name "guests"' % user.eid)
            user = cnx.user
            self.assertEqual(user.groups, set(('managers',)))
            cnx.commit()
            user = cnx.user
            self.assertEqual(user.groups, set(('managers', 'guests')))
            # cleanup
            cnx.execute('DELETE X in_group G WHERE X eid %s, G name "guests"' % user.eid)
            cnx.commit()

    def test_publish_validation_error(self):
        with self.admin_access.web_request() as req:
            user = req.user
            eid = text_type(user.eid)
            req.form = {
                'eid': eid,
                '__type:' + eid: 'CWUser',
                '_cw_entity_fields:' + eid: 'login-subject',
                'login-subject:' + eid: '',  # ERROR: no login specified
                # just a sample, missing some necessary information for real life
                '__errorurl': 'view?vid=edition...'
            }
            path, params = self.expect_redirect_handle_request(req, 'edit')
            forminfo = req.session.data['view?vid=edition...']
            eidmap = forminfo['eidmap']
            self.assertEqual(eidmap, {})
            values = forminfo['values']
            self.assertEqual(values['login-subject:' + eid], '')
            self.assertEqual(values['eid'], eid)
            error = forminfo['error']
            self.assertEqual(error.entity, user.eid)
            self.assertEqual(error.errors['login-subject'], 'required field')

    def test_validation_error_dont_loose_subentity_data_ctrl(self):
        """test creation of two linked entities

        error occurs on the web controller
        """
        with self.admin_access.web_request() as req:
            # set Y before X to ensure both entities are edited, not only X
            req.form = {'eid': ['Y', 'X'], '__maineid': 'X',
                        '__type:X': 'CWUser', '_cw_entity_fields:X': 'login-subject',
                        # missing required field
                        'login-subject:X': u'',
                        # but email address is set
                        '__type:Y': 'EmailAddress', '_cw_entity_fields:Y': 'address-subject',
                        'address-subject:Y': u'bougloup@logilab.fr',
                        'use_email-object:Y': 'X',
                        # necessary to get validation error handling
                        '__errorurl': 'view?vid=edition...',
                        }
            path, params = self.expect_redirect_handle_request(req, 'edit')
            forminfo = req.session.data['view?vid=edition...']
            self.assertEqual(set(forminfo['eidmap']), set('XY'))
            self.assertEqual(forminfo['eidmap']['X'], None)
            self.assertIsInstance(forminfo['eidmap']['Y'], int)
            self.assertEqual(forminfo['error'].entity, 'X')
            self.assertEqual(forminfo['error'].errors,
                             {'login-subject': 'required field'})
            self.assertEqual(forminfo['values'], req.form)

    def test_handle_request_with_lang_fromurl(self):
        """No language negociation, get language from URL."""
        self.config.global_set_option('language-mode', 'url-prefix')
        req = self.init_authentication('http')
        self.assertEqual(req.url(), 'http://testing.fr/cubicweb/login')
        self.assertEqual(req.lang, 'en')
        self.app.handle_request(req)
        newreq = self.requestcls(req.vreg, url='fr/toto')
        self.assertEqual(newreq.lang, 'en')
        self.assertEqual(newreq.url(), 'http://testing.fr/cubicweb/fr/toto')
        self.app.handle_request(newreq)
        self.assertEqual(newreq.lang, 'fr')
        self.assertEqual(newreq.url(), 'http://testing.fr/cubicweb/fr/toto')
        # unknown language
        newreq = self.requestcls(req.vreg, url='unknown-lang/cwuser')
        result = self.app.handle_request(newreq)
        self.assertEqual(newreq.lang, 'en')
        self.assertEqual(newreq.url(), 'http://testing.fr/cubicweb/unknown-lang/cwuser')
        self.assertIn('this resource does not exist',
                      result.decode('ascii', errors='ignore'))
        # no prefix
        newreq = self.requestcls(req.vreg, url='cwuser')
        result = self.app.handle_request(newreq)
        self.assertEqual(newreq.lang, 'en')
        self.assertEqual(newreq.url(), 'http://testing.fr/cubicweb/cwuser')
        self.assertNotIn('this resource does not exist',
                         result.decode('ascii', errors='ignore'))

    def test_handle_request_with_lang_negotiated(self):
        """Language negociated, normal case."""
        self.config.global_set_option('language-mode', 'http-negotiation')
        orig_translations = self.config.translations.copy()
        self.config.translations = {
            'fr': (text_type, lambda x, y: text_type(y)),
            'en': (text_type, lambda x, y: text_type(y))}
        try:
            headers = {'Accept-Language': 'fr'}
            with self.admin_access.web_request(headers=headers) as req:
                self.app.handle_request(req)
            self.assertEqual(req.lang, 'fr')
        finally:
            self.config.translations = orig_translations

    def test_handle_request_with_lang_negotiated_prefix_in_url(self):
        """Language negociated, unexpected language prefix in URL."""
        self.config.global_set_option('language-mode', 'http-negotiation')
        with self.admin_access.web_request(url='fr/toto') as req:
            result = self.app.handle_request(req)
        self.assertIn('this resource does not exist',  # NotFound.
                      result.decode('ascii', errors='ignore'))

    def test_handle_request_no_lang_negotiation_fixed_language(self):
        """No language negociation, "ui.language" fixed."""
        self.config.global_set_option('language-mode', '')
        vreg = self.app.vreg
        self.assertEqual(vreg.property_value('ui.language'), 'en')
        props = []
        try:
            with self.admin_access.cnx() as cnx:
                props.append(cnx.create_entity('CWProperty', value=u'de',
                                               pkey=u'ui.language').eid)
                cnx.commit()
            self.assertEqual(vreg.property_value('ui.language'), 'de')
            headers = {'Accept-Language': 'fr'}  # should not have any effect.
            with self.admin_access.web_request(headers=headers) as req:
                self.app.handle_request(req)
            # user has no "ui.language" property, getting site's default.
            self.assertEqual(req.lang, 'de')
            # XXX The following should work, but nasty handling of session and
            # request user make it fail...
            # with self.admin_access.cnx() as cnx:
            #     props.append(cnx.create_entity('CWProperty', value=u'es',
            #                                    pkey=u'ui.language',
            #                                    for_user=cnx.user).eid)
            #     cnx.commit()
            # with self.admin_access.web_request(headers=headers) as req:
            #     result = self.app.handle_request(req)
            # self.assertEqual(req.lang, 'es')
        finally:
            with self.admin_access.cnx() as cnx:
                for peid in props:
                    cnx.entity_from_eid(peid).cw_delete()
                cnx.commit()

    def test_validation_error_dont_loose_subentity_data_repo(self):
        """test creation of two linked entities

        error occurs on the repository
        """
        with self.admin_access.web_request() as req:
            # set Y before X to ensure both entities are edited, not only X
            req.form = {
                'eid': ['Y', 'X'], '__maineid': 'X',
                '__type:X': 'CWUser', '_cw_entity_fields:X': 'login-subject,upassword-subject',
                # already existent user
                'login-subject:X': u'admin',
                'upassword-subject:X': u'admin', 'upassword-subject-confirm:X': u'admin',
                '__type:Y': 'EmailAddress', '_cw_entity_fields:Y': 'address-subject',
                'address-subject:Y': u'bougloup@logilab.fr',
                'use_email-object:Y': 'X',
                # necessary to get validation error handling
                '__errorurl': 'view?vid=edition...',
            }
            path, params = self.expect_redirect_handle_request(req, 'edit')
            forminfo = req.session.data['view?vid=edition...']
            self.assertEqual(set(forminfo['eidmap']), set('XY'))
            self.assertIsInstance(forminfo['eidmap']['X'], int)
            self.assertIsInstance(forminfo['eidmap']['Y'], int)
            self.assertEqual(forminfo['error'].entity, forminfo['eidmap']['X'])
            expected_errors = {
                '': u'some relations violate a unicity constraint',
                'login': u'login is part of violated unicity constraint',
            }
            self.assertEqual(forminfo['error'].errors, expected_errors)
            self.assertEqual(forminfo['values'], req.form)

    def _edit_parent(self, dir_eid, parent_eid, role='subject',
                     etype='Directory', **kwargs):
        parent_eid = parent_eid or '__cubicweb_internal_field__'
        with self.admin_access.web_request() as req:
            req.form = {
                'eid': text_type(dir_eid),
                '__maineid': text_type(dir_eid),
                '__type:%s' % dir_eid: etype,
                'parent-%s:%s' % (role, dir_eid): parent_eid,
            }
            req.form.update(kwargs)
            req.form['_cw_entity_fields:%s' % dir_eid] = ','.join(
                ['parent-%s' % role] +
                [key.split(':')[0]
                 for key in kwargs.keys()
                 if not key.startswith('_')])
            self.expect_redirect_handle_request(req)

    def _edit_in_version(self, ticket_eid, version_eid, **kwargs):
        version_eid = version_eid or '__cubicweb_internal_field__'
        with self.admin_access.web_request() as req:
            req.form = {
                'eid': text_type(ticket_eid),
                '__maineid': text_type(ticket_eid),
                '__type:%s' % ticket_eid: 'Ticket',
                'in_version-subject:%s' % ticket_eid: version_eid,
            }
            req.form.update(kwargs)
            req.form['_cw_entity_fields:%s' % ticket_eid] = ','.join(
                ['in_version-subject'] +
                [key.split(':')[0]
                 for key in kwargs.keys()
                 if not key.startswith('_')])
            self.expect_redirect_handle_request(req)

    def test_create_and_link_directories(self):
        with self.admin_access.web_request() as req:
            req.form = {
                'eid': (u'A', u'B'),
                '__maineid': u'A',
                '__type:A': 'Directory',
                '__type:B': 'Directory',
                'parent-subject:B': u'A',
                'name-subject:A': u'topd',
                'name-subject:B': u'subd',
                '_cw_entity_fields:A': 'name-subject',
                '_cw_entity_fields:B': 'parent-subject,name-subject',
            }
            self.expect_redirect_handle_request(req)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', name=u'topd'))
            self.assertTrue(cnx.find('Directory', name=u'subd'))
            self.assertEqual(1, cnx.execute(
                'Directory SUBD WHERE SUBD parent TOPD,'
                ' SUBD name "subd", TOPD name "topd"').rowcount)

    def test_create_subentity(self):
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            cnx.commit()

        with self.admin_access.web_request() as req:
            req.form = {
                'eid': (text_type(topd.eid), u'B'),
                '__maineid': text_type(topd.eid),
                '__type:%s' % topd.eid: 'Directory',
                '__type:B': 'Directory',
                'parent-object:%s' % topd.eid: u'B',
                'name-subject:B': u'subd',
                '_cw_entity_fields:%s' % topd.eid: 'parent-object',
                '_cw_entity_fields:B': 'name-subject',
            }
            self.expect_redirect_handle_request(req)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', name=u'topd'))
            self.assertTrue(cnx.find('Directory', name=u'subd'))
            self.assertEqual(1, cnx.execute(
                'Directory SUBD WHERE SUBD parent TOPD,'
                ' SUBD name "subd", TOPD name "topd"').rowcount)

    def test_subject_subentity_removal(self):
        """Editcontroller: detaching a composite relation removes the subentity
        (edit from the subject side)
        """
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            sub1 = cnx.create_entity('Directory', name=u'sub1', parent=topd)
            sub2 = cnx.create_entity('Directory', name=u'sub2', parent=topd)
            cnx.commit()

        attrs = {'name-subject:%s' % sub1.eid: ''}
        self._edit_parent(sub1.eid, parent_eid=None, **attrs)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=topd.eid))
            self.assertFalse(cnx.find('Directory', eid=sub1.eid))
            self.assertTrue(cnx.find('Directory', eid=sub2.eid))

    def test_object_subentity_removal(self):
        """Editcontroller: detaching a composite relation removes the subentity
        (edit from the object side)
        """
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            sub1 = cnx.create_entity('Directory', name=u'sub1', parent=topd)
            sub2 = cnx.create_entity('Directory', name=u'sub2', parent=topd)
            cnx.commit()

        self._edit_parent(topd.eid, parent_eid=sub1.eid, role='object')

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=topd.eid))
            self.assertTrue(cnx.find('Directory', eid=sub1.eid))
            self.assertFalse(cnx.find('Directory', eid=sub2.eid))

    def test_reparent_subentity(self):
        "Editcontroller: re-parenting a subentity does not remove it"
        with self.admin_access.repo_cnx() as cnx:
            top1 = cnx.create_entity('Directory', name=u'top1')
            top2 = cnx.create_entity('Directory', name=u'top2')
            subd = cnx.create_entity('Directory', name=u'subd', parent=top1)
            cnx.commit()

        self._edit_parent(subd.eid, parent_eid=top2.eid)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=top1.eid))
            self.assertTrue(cnx.find('Directory', eid=top2.eid))
            self.assertTrue(cnx.find('Directory', eid=subd.eid))
            self.assertEqual(
                cnx.find('Directory', eid=subd.eid).one().parent[0], top2)

    def test_reparent_subentity_inlined(self):
        """Editcontroller: re-parenting a subentity does not remove it
        (inlined case)"""
        with self.admin_access.repo_cnx() as cnx:
            version1 = cnx.create_entity('Version', name=u'version1')
            version2 = cnx.create_entity('Version', name=u'version2')
            ticket = cnx.create_entity('Ticket', title=u'ticket',
                                       in_version=version1)
            cnx.commit()

        self._edit_in_version(ticket.eid, version_eid=version2.eid)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Version', eid=version1.eid))
            self.assertTrue(cnx.find('Version', eid=version2.eid))
            self.assertTrue(cnx.find('Ticket', eid=ticket.eid))
            self.assertEqual(
                cnx.find('Ticket', eid=ticket.eid).one().in_version[0], version2)

    def test_subject_mixed_composite_subentity_removal_1(self):
        """Editcontroller: detaching several subentities respects each rdef's
        compositeness - Remove non composite
        """
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            fs = cnx.create_entity('Filesystem', name=u'/tmp')
            subd = cnx.create_entity('Directory', name=u'subd',
                                     parent=(topd, fs))
            cnx.commit()

        self._edit_parent(subd.eid, parent_eid=topd.eid)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=topd.eid))
            self.assertTrue(cnx.find('Directory', eid=subd.eid))
            self.assertTrue(cnx.find('Filesystem', eid=fs.eid))
            self.assertEqual(cnx.find('Directory', eid=subd.eid).one().parent,
                             (topd,))

    def test_subject_mixed_composite_subentity_removal_2(self):
        """Editcontroller: detaching several subentities respects each rdef's
        compositeness - Remove composite
        """
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            fs = cnx.create_entity('Filesystem', name=u'/tmp')
            subd = cnx.create_entity('Directory', name=u'subd',
                                     parent=(topd, fs))
            cnx.commit()

        self._edit_parent(subd.eid, parent_eid=fs.eid)

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=topd.eid))
            self.assertFalse(cnx.find('Directory', eid=subd.eid))
            self.assertTrue(cnx.find('Filesystem', eid=fs.eid))

    def test_object_mixed_composite_subentity_removal_1(self):
        """Editcontroller: detaching several subentities respects each rdef's
        compositeness - Remove non composite
        """
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            fs = cnx.create_entity('Filesystem', name=u'/tmp')
            subd = cnx.create_entity('Directory', name=u'subd',
                                     parent=(topd, fs))
            cnx.commit()

        self._edit_parent(fs.eid, parent_eid=None, role='object',
                          etype='Filesystem')

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=topd.eid))
            self.assertTrue(cnx.find('Directory', eid=subd.eid))
            self.assertTrue(cnx.find('Filesystem', eid=fs.eid))
            self.assertEqual(cnx.find('Directory', eid=subd.eid).one().parent,
                             (topd,))

    def test_object_mixed_composite_subentity_removal_2(self):
        """Editcontroller: detaching several subentities respects each rdef's
        compositeness - Remove composite
        """
        with self.admin_access.repo_cnx() as cnx:
            topd = cnx.create_entity('Directory', name=u'topd')
            fs = cnx.create_entity('Filesystem', name=u'/tmp')
            subd = cnx.create_entity('Directory', name=u'subd',
                                     parent=(topd, fs))
            cnx.commit()

        self._edit_parent(topd.eid, parent_eid=None, role='object')

        with self.admin_access.repo_cnx() as cnx:
            self.assertTrue(cnx.find('Directory', eid=topd.eid))
            self.assertFalse(cnx.find('Directory', eid=subd.eid))
            self.assertTrue(cnx.find('Filesystem', eid=fs.eid))

    def test_delete_mandatory_composite(self):
        with self.admin_access.repo_cnx() as cnx:
            perm = cnx.create_entity('DirectoryPermission')
            mydir = cnx.create_entity('Directory', name=u'dir',
                                      has_permission=perm)
            cnx.commit()

        with self.admin_access.web_request() as req:
            dir_eid = text_type(mydir.eid)
            perm_eid = text_type(perm.eid)
            req.form = {
                'eid': [dir_eid, perm_eid],
                '__maineid': dir_eid,
                '__type:%s' % dir_eid: 'Directory',
                '__type:%s' % perm_eid: 'DirectoryPermission',
                '_cw_entity_fields:%s' % dir_eid: '',
                '_cw_entity_fields:%s' % perm_eid: 'has_permission-object',
                'has_permission-object:%s' % perm_eid: '',
            }
            path, _params = self.expect_redirect_handle_request(req, 'edit')
            self.assertTrue(req.find('Directory', eid=mydir.eid))
            self.assertFalse(req.find('DirectoryPermission', eid=perm.eid))

    def test_ajax_view_raise_arbitrary_error(self):
        class ErrorAjaxView(view.View):
            __regid__ = 'test.ajax.error'

            def call(self):
                raise Exception('whatever')

        with self.temporary_appobjects(ErrorAjaxView):
            with real_error_handling(self.app) as app:
                with self.admin_access.web_request(vid='test.ajax.error', url='') as req:
                    req.ajax_request = True
                    app.handle_request(req)
        self.assertEqual(http_client.INTERNAL_SERVER_ERROR,
                         req.status_out)

    def _test_cleaned(self, kwargs, injected, cleaned):
        with self.admin_access.web_request(**kwargs) as req:
            page = self.app_handle_request(req)
            self.assertNotIn(injected.encode('ascii'), page)
            self.assertIn(cleaned.encode('ascii'), page)

    def test_nonregr_script_kiddies(self):
        """test against current script injection"""
        injected = '<i>toto</i>'
        cleaned = 'toto'
        for kwargs in ({'vid': injected}, {'vtitle': injected}):
            with self.subTest(**kwargs):
                self._test_cleaned(kwargs, injected, cleaned)

    def test_site_wide_eproperties_sync(self):
        # XXX work in all-in-one configuration but not in twisted for instance
        # in which case we need a kindof repo -> http server notification
        # protocol
        vreg = self.app.vreg
        # default value
        self.assertEqual(vreg.property_value('ui.language'), 'en')
        with self.admin_access.client_cnx() as cnx:
            cnx.execute('INSERT CWProperty X: X value "fr", X pkey "ui.language"')
            self.assertEqual(vreg.property_value('ui.language'), 'en')
            cnx.commit()
            self.assertEqual(vreg.property_value('ui.language'), 'fr')
            cnx.execute('SET X value "de" WHERE X pkey "ui.language"')
            self.assertEqual(vreg.property_value('ui.language'), 'fr')
            cnx.commit()
            self.assertEqual(vreg.property_value('ui.language'), 'de')
            cnx.execute('DELETE CWProperty X WHERE X pkey "ui.language"')
            self.assertEqual(vreg.property_value('ui.language'), 'de')
            cnx.commit()
            self.assertEqual(vreg.property_value('ui.language'), 'en')

    # authentication tests ####################################################

    def test_http_auth_no_anon(self):
        req = self.init_authentication('http')
        self.assertAuthFailure(req)
        self.app.handle_request(req)
        self.assertEqual(401, req.status_out)
        clear_cache(req, 'get_authorization')
        authstr = base64.encodestring(('%s:%s' % (self.admlogin, self.admpassword)).encode('ascii'))
        req.set_request_header('Authorization', 'basic %s' % authstr.decode('ascii'))
        self.assertAuthSuccess(req)
        req._url = 'logout'
        self.assertRaises(LogOut, self.app_handle_request, req)
        self.assertEqual(len(self.open_sessions), 0)

    def test_cookie_auth_no_anon(self):
        req = self.init_authentication('cookie')
        self.assertAuthFailure(req)
        try:
            form = self.app.handle_request(req)
        except Redirect:
            self.fail('anonymous user should get login form')
        clear_cache(req, 'get_authorization')
        self.assertIn(b'__login', form)
        self.assertIn(b'__password', form)
        self.assertFalse(req.cnx)  # Mock cnx are False
        req.form['__login'] = self.admlogin
        req.form['__password'] = self.admpassword
        self.assertAuthSuccess(req)
        req._url = 'logout'
        self.assertRaises(LogOut, self.app_handle_request, req)
        self.assertEqual(len(self.open_sessions), 0)

    def test_login_by_email(self):
        with self.admin_access.client_cnx() as cnx:
            login = cnx.user.login
            address = login + u'@localhost'
            cnx.execute('INSERT EmailAddress X: X address %(address)s, U primary_email X '
                        'WHERE U login %(login)s', {'address': address, 'login': login})
            cnx.commit()
        req = self.init_authentication('cookie')
        self.set_option('allow-email-login', True)
        req.form['__login'] = address
        req.form['__password'] = self.admpassword
        self.assertAuthSuccess(req)
        req._url = 'logout'
        self.assertRaises(LogOut, self.app_handle_request, req)
        self.assertEqual(len(self.open_sessions), 0)

    def _reset_cookie(self, req):
        # preparing the suite of the test
        # set session id in cookie
        cookie = SimpleCookie()
        sessioncookie = self.app.session_handler.session_cookie(req)
        cookie[sessioncookie] = req.session.sessionid
        req.set_request_header('Cookie', cookie[sessioncookie].OutputString(),
                               raw=True)
        clear_cache(req, 'get_authorization')

    def _test_auth_anon(self, req):
        asession = self.app.get_session(req)
        # important otherwise _reset_cookie will not use the right session
        cnx = asession.new_cnx()
        with cnx:
            req.set_cnx(cnx)
        self.assertEqual(len(self.open_sessions), 1)
        self.assertEqual(asession.user.login, 'anon')
        self.assertTrue(asession.anonymous_session)
        self._reset_cookie(req)

    def _test_anon_auth_fail(self, req):
        self.assertEqual(1, len(self.open_sessions))
        session = self.app.get_session(req)
        cnx = session.new_cnx()
        with cnx:
            # important otherwise _reset_cookie will not use the right session
            req.set_cnx(cnx)
        self.assertEqual(req.message, 'authentication failure')
        self.assertEqual(req.session.anonymous_session, True)
        self.assertEqual(1, len(self.open_sessions))
        self._reset_cookie(req)

    def test_http_auth_anon_allowed(self):
        req = self.init_authentication('http', 'anon')
        self._test_auth_anon(req)
        authstr = base64.encodestring(b'toto:pouet')
        req.set_request_header('Authorization', 'basic %s' % authstr.decode('ascii'))
        self._test_anon_auth_fail(req)
        authstr = base64.encodestring(('%s:%s' % (self.admlogin, self.admpassword)).encode('ascii'))
        req.set_request_header('Authorization', 'basic %s' % authstr.decode('ascii'))
        self.assertAuthSuccess(req)
        req._url = 'logout'
        self.assertRaises(LogOut, self.app_handle_request, req)
        self.assertEqual(len(self.open_sessions), 0)

    def test_cookie_auth_anon_allowed(self):
        req = self.init_authentication('cookie', 'anon')
        self._test_auth_anon(req)
        req.form['__login'] = 'toto'
        req.form['__password'] = 'pouet'
        self._test_anon_auth_fail(req)
        req.form['__login'] = self.admlogin
        req.form['__password'] = self.admpassword
        self.assertAuthSuccess(req)
        req._url = 'logout'
        self.assertRaises(LogOut, self.app_handle_request, req)
        self.assertEqual(0, len(self.open_sessions))

    def test_anonymized_request(self):
        with self.admin_access.web_request() as req:
            self.assertEqual(self.admlogin, req.session.user.login)
            # admin should see anon + admin
            self.assertEqual(2, len(list(req.find('CWUser'))))
            with anonymized_request(req):
                self.assertEqual('anon', req.session.user.login)
                # anon should only see anon user
                self.assertEqual(1, len(list(req.find('CWUser'))))
            self.assertEqual(self.admlogin, req.session.user.login)
            self.assertEqual(2, len(list(req.find('CWUser'))))

    def test_non_regr_optional_first_var(self):
        with self.admin_access.web_request() as req:
            # expect a rset with None in [0][0]
            req.form['rql'] = 'rql:Any OV1, X WHERE X custom_workflow OV1?'
            self.app_handle_request(req)

    def test_handle_deprecation(self):
        """Test deprecation warning for *_handle methods."""
        with self.admin_access.web_request(url='foo') as req:
            with self.assertWarns(DeprecationWarning) as cm:
                self.app.core_handle(req, 'foo')
            self.assertIn('path argument got removed from "core_handle"',
                          str(cm.warning))
            with self.assertWarns(DeprecationWarning) as cm:
                self.app.main_handle_request('foo', req)
            self.assertIn('entry point arguments are now (req, path)',
                          str(cm.warning))


if __name__ == '__main__':
    unittest_main()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web import facet


class InsertAttrRelationTC(CubicWebTC):

    def parse(self, query):
        with self.admin_access.cnx() as cnx:
            rqlst = self.vreg.parse(cnx, query)
            rqlst.children[0]
        return rqlst

    def _generate(self, rqlst, rel, role, attr):
        select = rqlst.children[0]
        filtered_variable = facet.get_filtered_variable(select)
        facet.prepare_select(select, filtered_variable)
        facet.insert_attr_select_relation(select, filtered_variable,
                                          rel, role, attr)
        return rqlst.as_string()

    @property
    def select(self):
        return self.parse(u'Any B,(NOW - CD),S,V,U,GROUP_CONCAT(TN),VN,P,CD,BMD '
                           'GROUPBY B,CD,S,V,U,VN,P,BMD '
                           'WHERE B in_state S, B creation_date CD, '
                           'B modification_date BMD, T? tags B, T name TN, '
                           'V? bookmarked_by B, V title VN, B created_by U?, '
                           'B in_group P, P name "managers"')

    def test_1(self):
        self.assertEqual(self._generate(self.select, 'in_state', 'subject', 'name'),
                         'DISTINCT Any A,C ORDERBY C WHERE B in_group P, P name "managers", '
                         'B in_state A, B is CWUser, A name C')

    def test_2(self):
        self.assertEqual(self._generate(self.select, 'tags', 'object', 'name'),
                         'DISTINCT Any A,C ORDERBY C WHERE B in_group P, P name "managers", '
                         'A tags B, B is CWUser, A name C')

    def test_3(self):
        self.assertEqual(self._generate(self.select, 'created_by', 'subject', 'login'),
                         'DISTINCT Any A,C ORDERBY C WHERE B in_group P, P name "managers", '
                         'B created_by A, B is CWUser, A login C')

    def test_4(self):
        self.assertEqual(self._generate(self.parse(u'Any X WHERE X is CWUser'), 'created_by', 'subject', 'login'),
                          "DISTINCT Any A,B ORDERBY B WHERE X is CWUser, X created_by A, A login B")

    def test_5(self):
        self.assertEqual(self._generate(self.parse(u'Any X,L WHERE X is CWUser, X login L'), 'created_by', 'subject', 'login'),
                          "DISTINCT Any A,B ORDERBY B WHERE X is CWUser, X created_by A, A login B")

    def test_nonregr1(self):
        select = self.parse(u'Any T,V WHERE T bookmarked_by V?, '
                             'V in_state VS, VS name "published", T created_by U')
        self.assertEqual(self._generate(select, 'created_by', 'subject', 'login'),
                          "DISTINCT Any A,B ORDERBY B WHERE T created_by U, "
                          "T created_by A, T is Bookmark, A login B")

    def test_nonregr2(self):
        #'DISTINCT Any X,TMP,N WHERE P name TMP, X version_of P, P is Project, X is Version, not X in_state S,S name "published", X num N ORDERBY TMP,N'
        select = self.parse(u'DISTINCT Any V,TN,L ORDERBY TN,L WHERE T nom TN, V connait T, T is Personne, V is CWUser,'
                             'NOT V in_state VS, VS name "published", V login L')
        rschema = self.schema['connait']
        for rdefs in rschema.rdefs.values():
            rdefs.cardinality =  '++'
        try:
            self.assertEqual(self._generate(select, 'in_state', 'subject', 'name'),
                             'DISTINCT Any A,B ORDERBY B WHERE V is CWUser, '
                             'NOT EXISTS(V in_state VS), VS name "published", '
                             'V in_state A, A name B')
        finally:
            for rdefs in rschema.rdefs.values():
                rdefs.cardinality =  '**'

    def test_nonregr3(self):
        #'DISTINCT Any X,TMP,N WHERE P name TMP, X version_of P, P is Project, X is Version, not X in_state S,S name "published", X num N ORDERBY TMP,N'
        select = self.parse(u'DISTINCT Any X, MAX(Y) GROUPBY X WHERE X is CWUser, Y is Bookmark, X in_group A')
        self.assertEqual(self._generate(select, 'in_group', 'subject', 'name'),
                          "DISTINCT Any B,C ORDERBY C WHERE X is CWUser, X in_group B, B name C")


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""additional cubicweb-ctl commands and command handlers for cubicweb and
cubicweb's cubes development
"""
from __future__ import print_function

# *ctl module should limit the number of import to be imported as quickly as
# possible (for cubicweb-ctl reactivity, necessary for instance for usable bash
# completion). So import locally in command helpers.

import shutil
import tempfile
import sys
from datetime import datetime, date
from os import mkdir, chdir, path as osp
import pkg_resources
from warnings import warn

from pytz import UTC

from six.moves import input

from logilab.common import STD_BLACKLIST
from logilab.common.modutils import clean_sys_modules
from logilab.common.fileutils import ensure_fs_mode
from logilab.common.shellutils import find

from cubicweb.__pkginfo__ import version as cubicwebversion
from cubicweb import CW_SOFTWARE_ROOT as BASEDIR, BadCommandUsage, ExecutionError
from cubicweb.i18n import extract_from_tal, execute2
from cubicweb.cwctl import CWCTL
from cubicweb.cwconfig import CubicWebNoAppConfiguration
from cubicweb.toolsutils import (SKEL_EXCLUDE, Command, copy_skeleton,
                                 underline_title)
from cubicweb.web.webconfig import WebConfiguration
from cubicweb.server.serverconfig import ServerConfiguration


__docformat__ = "restructuredtext en"


STD_BLACKLIST = set(STD_BLACKLIST)
STD_BLACKLIST.add('.tox')
STD_BLACKLIST.add('test')
STD_BLACKLIST.add('node_modules')


class DevConfiguration(ServerConfiguration, WebConfiguration):
    """dummy config to get full library schema and appobjects for
    a cube or for cubicweb (without a home)
    """
    creating = True
    cleanup_unused_appobjects = False

    cubicweb_appobject_path = (ServerConfiguration.cubicweb_appobject_path
                               | WebConfiguration.cubicweb_appobject_path)
    cube_appobject_path = (ServerConfiguration.cube_appobject_path
                           | WebConfiguration.cube_appobject_path)

    def __init__(self, *cubes):
        super(DevConfiguration, self).__init__(cubes and cubes[0] or None)
        if cubes:
            self._cubes = self.reorder_cubes(
                self.expand_cubes(cubes, with_recommends=True))
            self.load_site_cubicweb()
        else:
            self._cubes = ()

    @property
    def apphome(self):
        return None

    def available_languages(self):
        return self.cw_languages()

    def main_config_file(self):
        return None
    def init_log(self):
        pass
    def load_configuration(self, **kw):
        pass
    def default_log_file(self):
        return None
    def default_stats_file(self):
        return None


def generate_schema_pot(w, cubedir=None):
    """generate a pot file with schema specific i18n messages

    notice that relation definitions description and static vocabulary
    should be marked using '_' and extracted using xgettext
    """
    from cubicweb.cwvreg import CWRegistryStore
    if cubedir:
        cube = osp.split(cubedir)[-1]
        if cube.startswith('cubicweb_'):
            cube = cube[len('cubicweb_'):]
        config = DevConfiguration(cube)
        depcubes = list(config._cubes)
        depcubes.remove(cube)
        libconfig = DevConfiguration(*depcubes)
    else:
        config = DevConfiguration()
        cube = libconfig = None
    clean_sys_modules(config.appobjects_modnames())
    schema = config.load_schema(remove_unused_rtypes=False)
    vreg = CWRegistryStore(config)
    # set_schema triggers objects registrations
    vreg.set_schema(schema)
    w(DEFAULT_POT_HEAD)
    _generate_schema_pot(w, vreg, schema, libconfig=libconfig)


def _generate_schema_pot(w, vreg, schema, libconfig=None):
    from cubicweb.i18n import add_msg
    from cubicweb.schema import NO_I18NCONTEXT, CONSTRAINTS
    w('# schema pot file, generated on %s\n'
      % datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S'))
    w('# \n')
    w('# singular and plural forms for each entity type\n')
    w('\n')
    vregdone = set()
    afss = vreg['uicfg']['autoform_section']
    aiams = vreg['uicfg']['actionbox_appearsin_addmenu']
    if libconfig is not None:
        # processing a cube, libconfig being a config with all its dependencies
        # (cubicweb incl.)
        from cubicweb.cwvreg import CWRegistryStore
        libschema = libconfig.load_schema(remove_unused_rtypes=False)
        clean_sys_modules(libconfig.appobjects_modnames())
        libvreg = CWRegistryStore(libconfig)
        libvreg.set_schema(libschema) # trigger objects registration
        libafss = libvreg['uicfg']['autoform_section']
        libaiams = libvreg['uicfg']['actionbox_appearsin_addmenu']
        # prefill vregdone set
        list(_iter_vreg_objids(libvreg, vregdone))

        def is_in_lib(rtags, eschema, rschema, role, tschema, predicate=bool):
            return any(predicate(rtag.etype_get(eschema, rschema, role, tschema))
                       for rtag in rtags)
    else:
        # processing cubicweb itself
        libschema = {}
        for cstrtype in CONSTRAINTS:
            add_msg(w, cstrtype)
        libafss = libaiams = None
        is_in_lib = lambda *args: False
    done = set()
    for eschema in sorted(schema.entities()):
        if eschema.type in libschema:
            done.add(eschema.description)
    for eschema in sorted(schema.entities()):
        etype = eschema.type
        if etype not in libschema:
            add_msg(w, etype)
            add_msg(w, '%s_plural' % etype)
            if not eschema.final:
                add_msg(w, 'This %s:' % etype)
                add_msg(w, 'New %s' % etype)
                add_msg(w, 'add a %s' % etype) # AddNewAction
                if libconfig is not None:  # processing a cube
                    # As of 3.20.3 we no longer use it, but keeping this string
                    # allows developers to run i18ncube with new cubicweb and still
                    # have the right translations at runtime for older versions
                    add_msg(w, 'This %s' % etype)
            if eschema.description and not eschema.description in done:
                done.add(eschema.description)
                add_msg(w, eschema.description)
        if eschema.final:
            continue
        for rschema, targetschemas, role in eschema.relation_definitions(True):
            if rschema.final:
                continue
            for tschema in targetschemas:

                for afs in afss:
                    fsections = afs.etype_get(eschema, rschema, role, tschema)
                    if 'main_inlined' in fsections and not \
                            is_in_lib(libafss, eschema, rschema, role, tschema,
                                      lambda x: 'main_inlined' in x):
                        add_msg(w, 'add a %s' % tschema,
                                'inlined:%s.%s.%s' % (etype, rschema, role))
                        add_msg(w, str(tschema),
                                'inlined:%s.%s.%s' % (etype, rschema, role))
                        break

                for aiam in aiams:
                    if aiam.etype_get(eschema, rschema, role, tschema) and not \
                            is_in_lib(libaiams, eschema, rschema, role, tschema):
                        if role == 'subject':
                            label = 'add %s %s %s %s' % (eschema, rschema,
                                                         tschema, role)
                            label2 = "creating %s (%s %%(linkto)s %s %s)" % (
                                tschema, eschema, rschema, tschema)
                        else:
                            label = 'add %s %s %s %s' % (tschema, rschema,
                                                         eschema, role)
                            label2 = "creating %s (%s %s %s %%(linkto)s)" % (
                                tschema, tschema, rschema, eschema)
                        add_msg(w, label)
                        add_msg(w, label2)
                        break
            # XXX also generate "creating ...' messages for actions in the
            # addrelated submenu
    w('# subject and object forms for each relation type\n')
    w('# (no object form for final or symmetric relation types)\n')
    w('\n')
    for rschema in sorted(schema.relations()):
        if rschema.type in libschema:
            done.add(rschema.type)
            done.add(rschema.description)
    for rschema in sorted(schema.relations()):
        rtype = rschema.type
        if rtype not in libschema:
            # bw compat, necessary until all translation of relation are done
            # properly...
            add_msg(w, rtype)
            done.add(rtype)
            if rschema.description and rschema.description not in done:
                add_msg(w, rschema.description)
            done.add(rschema.description)
            librschema = None
        else:
            librschema = libschema.rschema(rtype)
        # add context information only for non-metadata rtypes
        if rschema not in NO_I18NCONTEXT:
            libsubjects = librschema and librschema.subjects() or ()
            for subjschema in rschema.subjects():
                if not subjschema in libsubjects:
                    add_msg(w, rtype, subjschema.type)
        if not (rschema.final or rschema.symmetric):
            if rschema not in NO_I18NCONTEXT:
                libobjects = librschema and librschema.objects() or ()
                for objschema in rschema.objects():
                    if not objschema in libobjects:
                        add_msg(w, '%s_object' % rtype, objschema.type)
            if rtype not in libschema:
                # bw compat, necessary until all translation of relation are
                # done properly...
                add_msg(w, '%s_object' % rtype)
        for rdef in rschema.rdefs.values():
            if not rdef.description or rdef.description in done:
                continue
            if (librschema is None or
                (rdef.subject, rdef.object) not in librschema.rdefs or
                librschema.rdefs[(rdef.subject, rdef.object)].description != rdef.description):
                add_msg(w, rdef.description)
            done.add(rdef.description)
    for objid in _iter_vreg_objids(vreg, vregdone):
        add_msg(w, '%s_description' % objid)
        add_msg(w, objid)


def _iter_vreg_objids(vreg, done):
    for reg, objdict in vreg.items():
        if reg in ('boxes', 'contentnavigation'):
            continue
        for objects in objdict.values():
            for obj in objects:
                objid = '%s_%s' % (reg, obj.__regid__)
                if objid in done:
                    break
                pdefs = getattr(obj, 'cw_property_defs', {})
                if pdefs:
                    yield objid
                    done.add(objid)
                    break


DEFAULT_POT_HEAD = r'''msgid ""
msgstr ""
"Project-Id-Version: cubicweb %s\n"
"PO-Revision-Date: 2008-03-28 18:14+0100\n"
"Last-Translator: Logilab Team <contact@logilab.fr>\n"
"Language-Team: fr <contact@logilab.fr>\n"
"MIME-Version: 1.0\n"
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"Generated-By: cubicweb-devtools\n"
"Plural-Forms: nplurals=2; plural=(n > 1);\n"

''' % cubicwebversion


class UpdateCubicWebCatalogCommand(Command):
    """Update i18n catalogs for cubicweb library.

    It will regenerate cubicweb/i18n/xx.po files. You'll have then to edit those
    files to add translations of newly added messages.
    """
    name = 'i18ncubicweb'
    min_args = max_args = 0

    def run(self, args):
        """run the command with its specific arguments"""
        import shutil
        import tempfile
        import yams
        from logilab.common.fileutils import ensure_fs_mode
        from logilab.common.shellutils import globfind, find, rm
        from logilab.common.modutils import get_module_files
        from cubicweb.i18n import extract_from_tal, execute2
        tempdir = tempfile.mkdtemp(prefix='cw-')
        cwi18ndir = WebConfiguration.i18n_lib_dir()
        print('-> extract messages:', end=' ')
        print('schema', end=' ')
        schemapot = osp.join(tempdir, 'schema.pot')
        potfiles = [schemapot]
        potfiles.append(schemapot)
        # explicit close necessary else the file may not be yet flushed when
        # we'll using it below
        schemapotstream = open(schemapot, 'w')
        generate_schema_pot(schemapotstream.write, cubedir=None)
        schemapotstream.close()
        print('TAL', end=' ')
        tali18nfile = osp.join(tempdir, 'tali18n.py')
        extract_from_tal(find(osp.join(BASEDIR, 'web'), ('.py', '.pt')),
                         tali18nfile)
        print('-> generate .pot files.')
        pyfiles = get_module_files(BASEDIR)
        pyfiles += globfind(osp.join(BASEDIR, 'misc', 'migration'), '*.py')
        schemafiles = globfind(osp.join(BASEDIR, 'schemas'), '*.py')
        jsfiles = globfind(osp.join(BASEDIR, 'web'), 'cub*.js')
        for id, files, lang in [('pycubicweb', pyfiles, None),
                                ('schemadescr', schemafiles, None),
                                ('yams', get_module_files(yams.__path__[0]), None),
                                ('tal', [tali18nfile], None),
                                ('js', jsfiles, 'java'),
                                ]:
            potfile = osp.join(tempdir, '%s.pot' % id)
            cmd = ['xgettext', '--no-location', '--omit-header', '-k_']
            if lang is not None:
                cmd.extend(['-L', lang])
            cmd.extend(['-o', potfile])
            cmd.extend(files)
            execute2(cmd)
            if osp.exists(potfile):
                potfiles.append(potfile)
            else:
                print('-> WARNING: %s file was not generated' % potfile)
        print('-> merging %i .pot files' % len(potfiles))
        cubicwebpot = osp.join(tempdir, 'cubicweb.pot')
        cmd = ['msgcat', '-o', cubicwebpot] + potfiles
        execute2(cmd)
        print('-> merging main pot file with existing translations.')
        chdir(cwi18ndir)
        toedit = []
        for lang in CubicWebNoAppConfiguration.cw_languages():
            target = '%s.po' % lang
            cmd = ['msgmerge', '-N', '--sort-output', '-o',
                   target+'new', target, cubicwebpot]
            execute2(cmd)
            ensure_fs_mode(target)
            shutil.move('%snew' % target, target)
            toedit.append(osp.abspath(target))
        # cleanup
        rm(tempdir)
        # instructions pour la suite
        print('-> regenerated CubicWeb\'s .po catalogs.')
        print('\nYou can now edit the following files:')
        print('* ' + '\n* '.join(toedit))
        print('when you are done, run "cubicweb-ctl i18ncube yourcube".')


class UpdateCubeCatalogCommand(Command):
    """Update i18n catalogs for cubes. If no cube is specified, update
    catalogs of all registered cubes.
    """
    name = 'i18ncube'
    arguments = '[<cube>...]'

    def run(self, args):
        """run the command with its specific arguments"""
        if args:
            cubes = [DevConfiguration.cube_dir(cube) for cube in args]
        else:
            cubes = [DevConfiguration.cube_dir(cube)
                     for cube in DevConfiguration.available_cubes()]
            cubes = [cubepath for cubepath in cubes
                     if osp.exists(osp.join(cubepath, 'i18n'))]
        if not update_cubes_catalogs(cubes):
            raise ExecutionError("update cubes i18n catalog failed")


def update_cubes_catalogs(cubes):
    from subprocess import CalledProcessError
    for cubedir in cubes:
        if not osp.isdir(cubedir):
            print('-> ignoring %s that is not a directory.' % cubedir)
            continue
        try:
            toedit = update_cube_catalogs(cubedir)
        except CalledProcessError as exc:
            print('\n*** error while updating catalogs for cube', cubedir)
            print('cmd:\n%s' % exc.cmd)
            print('stdout:\n%s\nstderr:\n%s' % exc.data)
        except Exception:
            import traceback
            traceback.print_exc()
            print('*** error while updating catalogs for cube', cubedir)
            return False
        else:
            # instructions pour la suite
            if toedit:
                print('-> regenerated .po catalogs for cube %s.' % cubedir)
                print('\nYou can now edit the following files:')
                print('* ' + '\n* '.join(toedit))
                print ('When you are done, run "cubicweb-ctl i18ninstance '
                       '<yourinstance>" to see changes in your instances.')
            return True


class I18nCubeMessageExtractor(object):
    """This class encapsulates all the xgettext extraction logic

    ``generate_pot_file`` is the main entry point called by the ``i18ncube``
    command. A cube might decide to customize extractors to ignore a given
    directory or to extract messages from a new file type (e.g. .jinja2 files)

    For each file type, the class must define two methods:

    - ``collect_{filetype}()`` that must return the list of files
      xgettext should inspect,

    - ``extract_{filetype}(files)`` that calls xgettext and returns the
      path to the generated ``pot`` file
    """
    blacklist = STD_BLACKLIST
    formats = ['tal', 'js', 'py']

    def __init__(self, workdir, cubedir):
        self.workdir = workdir
        self.cubedir = cubedir

    def generate_pot_file(self):
        """main entry point: return the generated ``cube.pot`` file

        This function first generates all the pot files (schema, tal,
        py, js) and then merges them in a single ``cube.pot`` that will
        be used to eventually update the ``i18n/*.po`` files.
        """
        potfiles = self.generate_pot_files()
        potfile = osp.join(self.workdir, 'cube.pot')
        print('-> merging %i .pot files' % len(potfiles))
        cmd = ['msgcat', '-o', potfile]
        cmd.extend(potfiles)
        execute2(cmd)
        return potfile if osp.exists(potfile) else None

    def find(self, exts, blacklist=None):
        """collect files with extensions ``exts`` in the cube directory
        """
        if blacklist is None:
            blacklist = self.blacklist
        return find(self.cubedir, exts, blacklist=blacklist)

    def generate_pot_files(self):
        """generate and return the list of all ``pot`` files for the cube

        - static-messages.pot,
        - schema.pot,
        - one ``pot`` file for each inspected format (.py, .js, etc.)
        """
        print('-> extracting messages:', end=' ')
        potfiles = []
        # static messages
        if osp.exists(osp.join('i18n', 'entities.pot')):
            warn('entities.pot is deprecated, rename file '
                 'to static-messages.pot (%s)'
                 % osp.join('i18n', 'entities.pot'), DeprecationWarning)
            potfiles.append(osp.join('i18n', 'entities.pot'))
        elif osp.exists(osp.join('i18n', 'static-messages.pot')):
            potfiles.append(osp.join('i18n', 'static-messages.pot'))
        # messages from schema
        potfiles.append(self.schemapot())
        # messages from sourcecode
        for fmt in self.formats:
            collector = getattr(self, 'collect_{0}'.format(fmt))
            extractor = getattr(self, 'extract_{0}'.format(fmt))
            files = collector()
            if files:
                potfile = extractor(files)
                if potfile:
                    potfiles.append(potfile)
        return potfiles

    def schemapot(self):
        """generate the ``schema.pot`` file"""
        schemapot = osp.join(self.workdir, 'schema.pot')
        print('schema', end=' ')
        # explicit close necessary else the file may not be yet flushed when
        # we'll using it below
        schemapotstream = open(schemapot, 'w')
        generate_schema_pot(schemapotstream.write, self.cubedir)
        schemapotstream.close()
        return schemapot

    def _xgettext(self, files, output, k='_', extraopts=''):
        """shortcut to execute the xgettext command and return output file
        """
        tmppotfile = osp.join(self.workdir, output)
        cmd = ['xgettext', '--no-location', '--omit-header', '-k' + k,
               '-o', tmppotfile] + extraopts.split() + files
        execute2(cmd)
        if osp.exists(tmppotfile):
            return tmppotfile

    def collect_tal(self):
        print('TAL', end=' ')
        return self.find(('.py', '.pt'))

    def extract_tal(self, files):
        tali18nfile = osp.join(self.workdir, 'tali18n.py')
        extract_from_tal(files, tali18nfile)
        return self._xgettext(files, output='tal.pot')

    def collect_js(self):
        print('Javascript')
        return [jsfile for jsfile in self.find('.js')
                if osp.basename(jsfile).startswith('cub')]

    def extract_js(self, files):
        return self._xgettext(files, output='js.pot',
                              extraopts='-L java --from-code=utf-8')

    def collect_py(self):
        print('-> creating cube-specific catalog')
        return self.find('.py')

    def extract_py(self, files):
        return self._xgettext(files, output='py.pot')


def update_cube_catalogs(cubedir):
    cubedir = osp.abspath(osp.normpath(cubedir))
    workdir = tempfile.mkdtemp()
    try:
        cubename = osp.basename(cubedir)
        if cubename.startswith('cubicweb_'):  # new layout
            distname = cubename
            cubename = cubename[len('cubicweb_'):]
        else:
            distname = 'cubicweb_' + cubename
        print('cubedir', cubedir)
        extract_cls = I18nCubeMessageExtractor
        try:
            extract_cls = pkg_resources.load_entry_point(
                distname, 'cubicweb.i18ncube', cubename)
        except (pkg_resources.DistributionNotFound, ImportError):
            pass  # no customization found
        print(underline_title('Updating i18n catalogs for cube %s' % cubename))
        chdir(cubedir)
        extractor = extract_cls(workdir, cubedir)
        potfile = extractor.generate_pot_file()
        if potfile is None:
            print('no message catalog for cube', cubename, 'nothing to translate')
            return ()
        print('-> merging main pot file with existing translations:', end=' ')
        chdir('i18n')
        toedit = []
        for lang in CubicWebNoAppConfiguration.cw_languages():
            print(lang, end=' ')
            cubepo = '%s.po' % lang
            if not osp.exists(cubepo):
                shutil.copy(potfile, cubepo)
            else:
                cmd = ['msgmerge', '-N', '-s', '-o', cubepo + 'new',
                       cubepo, potfile]
                execute2(cmd)
                ensure_fs_mode(cubepo)
                shutil.move('%snew' % cubepo, cubepo)
            toedit.append(osp.abspath(cubepo))
        print()
        return toedit
    finally:
        # cleanup
        shutil.rmtree(workdir)


class NewCubeCommand(Command):
    """Create a new cube.

    <cubename>
      the name of the new cube. It should be a valid python module name.
    """
    name = 'newcube'
    arguments = '<cubename>'
    min_args = max_args = 1
    options = (
        ("layout",
         {'short': 'L', 'type' : 'choice', 'metavar': '<cube layout>',
          'default': 'simple', 'choices': ('simple', 'full'),
          'help': 'cube layout. You\'ll get a minimal cube with the "simple" \
layout, and a full featured cube with "full" layout.',
          }
         ),
        ("directory",
         {'short': 'd', 'type' : 'string', 'metavar': '<cubes directory>',
          'help': 'directory where the new cube should be created',
          }
         ),
        ("verbose",
         {'short': 'v', 'type' : 'yn', 'metavar': '<verbose>',
          'default': 'n',
          'help': 'verbose mode: will ask all possible configuration questions',
          }
         ),
        ("author",
         {'short': 'a', 'type' : 'string', 'metavar': '<author>',
          'default': 'LOGILAB S.A. (Paris, FRANCE)',
          'help': 'cube author',
          }
         ),
        ("author-email",
         {'short': 'e', 'type' : 'string', 'metavar': '<email>',
          'default': 'contact@logilab.fr',
          'help': 'cube author\'s email',
          }
         ),
        ("author-web-site",
         {'short': 'w', 'type' : 'string', 'metavar': '<web site>',
          'default': 'http://www.logilab.fr',
          'help': 'cube author\'s web site',
          }
         ),
        ("license",
         {'short': 'l', 'type' : 'choice', 'metavar': '<license>',
          'default': 'LGPL', 'choices': ('GPL', 'LGPL', ''),
          'help': 'cube license',
          }
         ),
        )

    LICENSES = {
        'LGPL': '''\
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
''',

        'GPL': '''\
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 2.1 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
''',
        '': '# INSERT LICENSE HERE'
        }

    def run(self, args):
        import re
        from logilab.common.shellutils import ASK
        cubename = args[0]
        if not re.match('[_A-Za-z][_A-Za-z0-9]*$', cubename):
            raise BadCommandUsage(
                'cube name must be a valid python module name')
        verbose = self.get('verbose')
        destdir = self.get('directory')
        if not destdir:
            cubespath = ServerConfiguration.cubes_search_path()
            if len(cubespath) > 1:
                raise BadCommandUsage(
                    "can't guess directory where to put the new cube."
                    " Please specify it using the --directory option")
            destdir = cubespath[0]
        if not osp.isdir(destdir):
            print("-> creating cubes directory", destdir)
            try:
                mkdir(destdir)
            except OSError as err:
                self.fail("failed to create directory %r\n(%s)"
                          % (destdir, err))
        default_name = 'cubicweb-%s' % cubename.lower().replace('_', '-')
        if verbose:
            distname = input('Debian name for your cube ? [%s]): '
                             % default_name).strip()
            if not distname:
                distname = default_name
            elif not distname.startswith('cubicweb-'):
                if ASK.confirm('Do you mean cubicweb-%s ?' % distname):
                    distname = 'cubicweb-' + distname
        else:
            distname = default_name
        if not re.match('[a-z][-a-z0-9]*$', distname):
            raise BadCommandUsage(
                'cube distname should be a valid debian package name')
        cubedir = osp.join(destdir, distname)
        if osp.exists(cubedir):
            self.fail("%s already exists!" % cubedir)
        skeldir = osp.join(BASEDIR, 'skeleton')
        longdesc = shortdesc = input(
            'Enter a short description for your cube: ')
        if verbose:
            longdesc = input(
                'Enter a long description (leave empty to reuse the short one): ')
        dependencies = {
            'six': '>= 1.4.0',
            'cubicweb': '>= %s' % cubicwebversion,
        }
        if verbose:
            dependencies.update(self._ask_for_dependencies())
        context = {
            'cubename': cubename,
            'distname': distname,
            'shortdesc': shortdesc,
            'longdesc': longdesc or shortdesc,
            'dependencies': dependencies,
            'version': cubicwebversion,
            'year': str(date.today().year),
            'author': self['author'],
            'author-email': self['author-email'],
            'rfc2822-date': datetime.now(tz=UTC).strftime('%a, %d %b %Y %T %z'),
            'author-web-site': self['author-web-site'],
            'license': self['license'],
            'long-license': self.LICENSES[self['license']],
        }
        exclude = SKEL_EXCLUDE
        if self['layout'] == 'simple':
            exclude += ('sobjects.py*', 'precreate.py*', 'realdb_test*',
                        'cubes.*', 'uiprops.py*')
        copy_skeleton(skeldir, cubedir, context, exclude=exclude)

    def _ask_for_dependencies(self):
        from logilab.common.shellutils import ASK
        from logilab.common.textutils import splitstrip
        depcubes = []
        for cube in ServerConfiguration.available_cubes():
            answer = ASK.ask("Depends on cube %s? " % cube,
                             ('N','y','skip','type'), 'N')
            if answer == 'y':
                depcubes.append(cube)
            if answer == 'type':
                depcubes = splitstrip(input('type dependencies: '))
                break
            elif answer == 'skip':
                break
        return dict(('cubicweb-' + cube, ServerConfiguration.cube_version(cube))
                    for cube in depcubes)


class ExamineLogCommand(Command):
    """Examine a rql log file.

    Will print out the following table

      Percentage; Cumulative Time (clock); Cumulative Time (CPU); Occurences; Query

    sorted by descending cumulative time (clock). Time are expressed in seconds.

    Chances are the lines at the top are the ones that will bring the higher
    benefit after optimisation. Start there.
    """
    arguments = 'rql.log'
    name = 'exlog'
    options = ()

    def run(self, args):
        import re
        requests = {}
        for filepath in args:
            try:
                stream = open(filepath)
            except OSError as ex:
                raise BadCommandUsage("can't open rql log file %s: %s"
                                      % (filepath, ex))
            for lineno, line in enumerate(stream):
                if not ' WHERE ' in line:
                    continue
                try:
                    rql, time = line.split('--')
                    rql = re.sub("(\'\w+': \d*)", '', rql)
                    if '{' in rql:
                        rql = rql[:rql.index('{')]
                    req = requests.setdefault(rql, [])
                    time.strip()
                    chunks = time.split()
                    clocktime = float(chunks[0][1:])
                    cputime = float(chunks[-3])
                    req.append( (clocktime, cputime) )
                except Exception as exc:
                    sys.stderr.write('Line %s: %s (%s)\n' % (lineno, exc, line))
        stat = []
        for rql, times in requests.items():
            stat.append( (sum(time[0] for time in times),
                          sum(time[1] for time in times),
                          len(times), rql) )
        stat.sort()
        stat.reverse()
        total_time = sum(clocktime for clocktime, cputime, occ, rql in stat) * 0.01
        print('Percentage;Cumulative Time (clock);Cumulative Time (CPU);Occurences;Query')
        for clocktime, cputime, occ, rql in stat:
            print('%.2f;%.2f;%.2f;%s;%s' % (clocktime/total_time, clocktime,
                                            cputime, occ, rql))


class GenerateSchema(Command):
    """Generate schema image for the given cube"""
    name = "schema"
    arguments = '<cube>'
    min_args = max_args = 1
    options = [
        ('output-file',
         {'type':'string', 'default': None,
          'metavar': '<file>', 'short':'o', 'help':'output image file',
          'input':False,
          }),
        ('viewer',
         {'type': 'string', 'default':None,
          'short': "d", 'metavar':'<cmd>',
          'help':'command use to view the generated file (empty for none)',
          }),
        ('show-meta',
         {'action': 'store_true', 'default':False,
          'short': "m", 'metavar': "<yN>",
          'help':'include meta and internal entities in schema',
          }),
        ('show-workflow',
         {'action': 'store_true', 'default':False,
          'short': "w", 'metavar': "<yN>",
          'help':'include workflow entities in schema',
          }),
        ('show-cw-user',
         {'action': 'store_true', 'default':False,
          'metavar': "<yN>",
          'help':'include cubicweb user entities in schema',
          }),
        ('exclude-type',
         {'type':'string', 'default':'',
          'short': "x", 'metavar': "<types>",
          'help':'coma separated list of entity types to remove from view',
          }),
        ('include-type',
         {'type':'string', 'default':'',
          'short': "i", 'metavar': "<types>",
          'help':'coma separated list of entity types to include in view',
          }),
        ('show-etype',
         {'type':'string', 'default':'',
          'metavar': '<etype>',
          'help':'show graph of this etype and its neighbours'
          }),
        ]

    def run(self, args):
        from subprocess import Popen
        from tempfile import NamedTemporaryFile
        from logilab.common.textutils import splitstrip
        from logilab.common.graph import GraphGenerator, DotBackend
        from yams import schema2dot as s2d, BASE_TYPES
        from cubicweb.schema import (META_RTYPES, SCHEMA_TYPES, SYSTEM_RTYPES,
                                     WORKFLOW_TYPES, INTERNAL_TYPES)
        cubes = splitstrip(args[0])
        dev_conf = DevConfiguration(*cubes)
        schema = dev_conf.load_schema()
        out, viewer = self['output-file'], self['viewer']
        if out is None:
            tmp_file = NamedTemporaryFile(suffix=".svg")
            out = tmp_file.name
        skiptypes = BASE_TYPES | SCHEMA_TYPES
        if not self['show-meta']:
            skiptypes |=  META_RTYPES | SYSTEM_RTYPES | INTERNAL_TYPES
        if not self['show-workflow']:
            skiptypes |= WORKFLOW_TYPES
        if not self['show-cw-user']:
            skiptypes |= set(('CWUser', 'CWGroup', 'EmailAddress'))
        skiptypes |= set(self['exclude-type'].split(','))
        skiptypes -= set(self['include-type'].split(','))

        if not self['show-etype']:
            s2d.schema2dot(schema, out, skiptypes=skiptypes)
        else:
            etype = self['show-etype']
            visitor = s2d.OneHopESchemaVisitor(schema[etype], skiptypes=skiptypes)
            propshdlr = s2d.SchemaDotPropsHandler(visitor)
            backend = DotBackend('schema', 'BT',
                                 ratio='compress',size=None,
                                 renderer='dot',
                                 additionnal_param={'overlap' : 'false',
                                                    'splines' : 'true',
                                                    'sep' : '0.2'})
            generator = s2d.GraphGenerator(backend)
            generator.generate(visitor, propshdlr, out)

        if viewer:
            p = Popen((viewer, out))
            p.wait()


for cmdcls in (UpdateCubicWebCatalogCommand,
               UpdateCubeCatalogCommand,
               NewCubeCommand,
               ExamineLogCommand,
               GenerateSchema,
               ):
    CWCTL.register(cmdcls)
